"""
Run tests of disease models
"""

# %% Imports and settings
import sciris as sc
import matplotlib.pyplot as plt
import starsim as ss
import starsim_examples as sse

test_run = True
n_agents = [10_000, 2_000][test_run]
do_plot = True
sc.options(interactive=do_plot) # Assume not running interactively


@sc.timer()
def test_sir():
    sc.heading('Testing SIR dynamics')

    ppl = ss.People(n_agents)
    network_pars = {
        'n_contacts': ss.poisson(4), # Contacts Poisson distributed with a mean of 4
    }
    networks = ss.RandomNet(**network_pars)

    sir_pars = {
        'dur_inf': ss.normal(loc=10),  # Override the default distribution
    }
    sir = ss.SIR(**sir_pars)

    # Change pars after creating the SIR instance
    sir.pars.beta = {'random': ss.freqperyear(0.1)}

    # You can also change the parameters of the default lognormal distribution directly
    sir.pars.dur_inf.set(loc=5)

    # Or use a function, here a lambda that makes the mean for each agent equal to their age divided by 10
    sir.pars.dur_inf.set(loc = lambda self, sim, uids: sim.people.age[uids] / 10)

    sim = ss.Sim(people=ppl, diseases=sir, networks=networks)
    sim.run()

    plt.figure()
    res = sim.results
    plt.stackplot(
        sim.timevec,
        res.sir.n_susceptible,
        res.sir.n_infected,
        res.sir.n_recovered,
        res.new_deaths.cumsum(),
    )
    plt.legend(['Susceptible', 'Infected', 'Recovered', 'Dead'])
    plt.xlabel('Year')
    plt.title('SIR')
    return sim


@sc.timer()
def test_sir_epi():
    sc.heading('Test basic epi dynamics')

    base_pars = dict(n_agents=n_agents, networks=dict(type='random'), diseases=dict(type='sir'))

    # Define the parameters to vary
    par_effects = dict(
        beta = [0.01, 0.99],
        n_contacts = [1, 20],
        init_prev = [0.1, 0.9],
        dur_inf = [1, 8],
        p_death = [.01, .1],
    )

    # Loop over each of the above parameters and make sure they affect the epi dynamics in the expected ways
    for par, par_val in par_effects.items():
        lo = par_val[0]
        hi = par_val[1]

        # Make baseline pars
        pars0 = sc.dcp(base_pars)
        pars1 = sc.dcp(base_pars)

        if par != 'n_contacts':
            pars0['diseases'] = sc.mergedicts(pars0['diseases'], {par: lo})
            pars1['diseases'] = sc.mergedicts(pars1['diseases'], {par: hi})
        else:
            pars0['networks'] = sc.mergedicts(pars0['networks'], {par: lo})
            pars1['networks'] = sc.mergedicts(pars1['networks'], {par: hi})

        # Run the simulations and pull out the results
        s0 = ss.Sim(pars0, label=f'{par} {par_val[0]}').run()
        s1 = ss.Sim(pars1, label=f'{par} {par_val[1]}').run()

        # Check results
        if par == 'p_death':
            v0 = s0.results.cum_deaths[-1]
            v1 = s1.results.cum_deaths[-1]
        else:
            ind = 1 if par == 'init_prev' else -1
            v0 = s0.results.sir.prevalence[ind]
            v1 = s1.results.sir.prevalence[ind]

        print(f'Checking with varying {par:10s} ... ', end='')
        assert v0 <= v1, f'Expected prevalence to be lower with {par}={lo} than with {par}={hi}, but {v0} > {v1})'
        print(f'✓ ({v0} <= {v1})')

    return s0, s1


@sc.timer()
def test_sis(do_plot=do_plot):
    sc.heading('Testing SIS dynamics')

    pars = dict(
        n_agents = n_agents,
        diseases = 'sis',
        networks = 'random'
    )
    sim = ss.Sim(pars)
    sim.run()
    if do_plot:
        sim.plot()
        sim.diseases.sis.plot()
    return sim


@sc.timer()
def test_ncd():
    sc.heading('Testing NCDs')

    ppl = ss.People(n_agents)
    ncd = ss.NCD()
    sim = ss.Sim(people=ppl, diseases=ncd, copy_inputs=False, dt=ss.years(1), analyzers='infection_log') # Since using ncd directly below
    sim.run()
    log = sim.analyzers[0].logs[0]

    assert len(log.out_edges) == log.number_of_edges()
    df = log.to_df()  # Check generation of line-list
    assert df.source.isna().all()

    plt.figure()
    plt.stackplot(
        sim.timevec,
        ncd.results.n_not_at_risk,
        ncd.results.n_at_risk - ncd.results.n_affected,
        ncd.results.n_affected,
        sim.results.new_deaths.cumsum(),
    )
    assert ncd.results.n_not_at_risk.label == 'Not at risk'
    assert ncd.results.n_affected.label == 'Affected'
    plt.legend(['Not at risk', 'At risk', 'Affected', 'Dead'])
    plt.xlabel('Year')
    plt.title('NCD')
    return sim


@sc.timer()
def test_gavi():
    sc.heading('Testing GAVI diseases')
    ss.register_modules(sse) # So we can use strings for these

    sims = sc.autolist()
    for disease in ['cholera', 'measles', 'ebola']:
        pars = dict(
            diseases = disease,
            n_agents = n_agents,
            networks = 'random',
        )
        sim = ss.Sim(pars)
        sim.run()
        sim.plot()
        sims += sim
    return sims


@sc.timer()
def test_multidisease():
    sc.heading('Testing simulating multiple diseases')

    ppl = ss.People(n_agents)
    sir1 = ss.SIR(name='sir1')
    sir2 = ss.SIR(name='sir2')

    sir1.pars.beta = {'randomnet': 0.1}
    sir2.pars.beta = {'randomnet': 0.2}
    networks = ss.RandomNet(n_contacts=ss.poisson(4))

    sim = ss.Sim(people=ppl, diseases=[sir1, sir2], networks=networks)
    sim.run()
    return sim


@sc.timer()
def test_mtct():
    sc.heading('Test mother-to-child transmission routes')
    ppl = ss.People(n_agents)
    sis = ss.SIS(
        beta = dict(
            random = [ss.permonth(0.005)]*2,
            prenatal = [ss.permonth(0.1), 0],
            postnatal = [ss.permonth(0.1), 0]
        )
    )
    networks = [ss.RandomNet(), ss.PrenatalNet(), ss.PostnatalNet()]
    demographics = ss.Pregnancy(fertility_rate=ss.freqperyear(20))
    sim = ss.Sim(dt=ss.month, people=ppl, diseases=sis, networks=networks, demographics=demographics)
    sim.run()
    sim.plot()
    return sim


if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)
    sir   = test_sir()
    s1,s2 = test_sir_epi()
    sis   = test_sis()
    ncd   = test_ncd()
    gavi  = test_gavi()
    multi = test_multidisease()
    mtct  = test_mtct()



#!/usr/bin/env python3
"""
Check that the migration script examples work (written mostly by ChatGPT)
"""

import shutil
import tempfile
import difflib
import sciris as sc

root = sc.path(".")

# Find all base names that have __script.py, which should have matching __v2.py and __v3.py files
triplets = []
for script_file in sorted(root.rglob("*__script.py")):
    base = script_file.stem.replace("__script", "")
    input_file = script_file.with_name(f"{base}__v2.py")
    output_file = input_file.with_name(f"{base}__v3.py")
    if input_file.exists() and output_file.exists():
        triplets.append((base, input_file, output_file, script_file))
    else:
        raise FileNotFoundError(f'Expecting {input_file} and {output_file}')

for i, (base, input_file, output_file, script_file) in enumerate(triplets):
    expected = output_file.read_text().splitlines(keepends=True)
    with tempfile.TemporaryDirectory() as tmpdir:
        print(sc.ansi.green(f"Checking example {base} ({i+1} of {len(triplets)}) in {tmpdir}/..."))

        tmpdir = sc.path(tmpdir)
        tmp_example = tmpdir / f"{base}__example.py"
        tmp_script = tmpdir / f"{base}__script.py"

        # Copy files
        shutil.copy(input_file, tmp_example)
        shutil.copy(script_file, tmp_script)

        # Run the script
        cmd = f'python {tmp_script} {tmp_example}'
        sc.runcommand(cmd, cwd=tmpdir)

        # Compare the transformed input with the expected output
        actual = tmp_example.read_text().splitlines(keepends=True)
        if actual == expected:
            print(f"  ✅ {base} matches expected output.")
        else:
            print(f"  ❌ {base} differs from expected output.")
            diff = "".join(difflib.unified_diff(actual, expected, fromfile="actual", tofile="expected"))
            print(diff)
            print('\n\nRaw strings:')
            print('<<<')
            print(f'"{expected}"')
            print('≠≠≠≠≠≠≠≠≠')
            print(f'"{actual}"')
            print('>>>')



"""
Define pregnancy, deaths, migration, etc.
"""
import numpy as np
import starsim as ss
import sciris as sc
import pandas as pd

ss_float = ss.dtypes.float
ss_int = ss.dtypes.int
_ = None

__all__ = ['Demographics', 'Births', 'Deaths', 'Pregnancy']


class Demographics(ss.Module):
    """
    A demographic module typically handles births/deaths/migration and takes
    place at the start of the timestep, before networks are updated and before
    any disease modules are executed.
    """
    pass


class Births(Demographics):
    """
    Create births based on rates, rather than based on pregnancy.

    Births are generated using population-level birth rates that can vary
    by year. The number of births per timestep is determined by applying
    the birth rate to the current population size.

    Args:
        birth_rate (float/rate/dataframe): value for birth rate, or birth rate data
        rel_birth (float): constant used to scale all birth rates
        rate_units (float): units for birth rates (default assumes per 1000)
        metadata (dict): dict with data column mappings for birth rate data (if birth_rate is a dataframe)
    """
    def __init__(self, pars=None, birth_rate=_, rel_birth=_, rate_units=_, metadata=None, **kwargs):
        
        super().__init__()
        self.define_pars(
            birth_rate = ss.peryear(20),
            rel_birth = 1,
            rate_units = 1e-3,  # assumes birth rates are per 1000. If using percentages, switch this to 1
        )
        self.update_pars(pars, **kwargs)

        # Process metadata. Defaults here are the labels used by UN data
        self.metadata = sc.mergedicts(
            sc.objdict(data_cols=dict(year='Year', value='CBR')),
            metadata,
        )

        # Process data, which may be provided as a number, dict, dataframe, or series
        # If it's a number it's left as-is; otherwise it's converted to a dataframe
        self.pars.birth_rate = self.standardize_birth_data()
        self.n_births = 0 # For results tracking
        self.dist = ss.bernoulli(p=0) # Used to generate random numbers; set to use the birth rate above
        return

    def init_pre(self, sim):
        """ Initialize with sim information """
        super().init_pre(sim)
        if isinstance(self.pars.birth_rate, pd.DataFrame):
            br_year = self.pars.birth_rate[self.metadata.data_cols['year']]
            br_val = self.pars.birth_rate[self.metadata.data_cols['cbr']]
            all_birth_rates = np.interp(self.timevec, br_year, br_val) # This assumes a year timestep -- probably ok?
            self.pars.birth_rate = all_birth_rates
        return

    def standardize_birth_data(self):
        """ Standardize/validate birth rates - handled in an external file due to shared functionality """
        birth_rate = ss.standardize_data(data=self.pars.birth_rate, metadata=self.metadata)
        if isinstance(birth_rate, (pd.Series, pd.DataFrame)):
            return birth_rate.xs(0,level='age')
        if sc.isnumber(birth_rate):
            # If the user has provided a bare number, assume it is per year
            msg = f'Birth rate was specified as a number rather than a rate - assuming it is {birth_rate} per year'
            ss.warn(msg)
            return ss.peryear(birth_rate)
        return birth_rate

    def init_results(self):
        super().init_results()
        self.define_results(
            ss.Result('new',        dtype=int,   scale=True,  summarize_by='sum',  label='New births'),
            ss.Result('cumulative', dtype=int,   scale=True,  summarize_by='last', label='Cumulative births'),
            ss.Result('cbr',        dtype=float, scale=False, summarize_by='mean', label='Crude birth rate'),
        )
        return

    def get_births(self):
        """
        Extract the right birth rates to use and translate it into a number of people to add.
        """
        sim = self.sim
        p = self.pars

        if isinstance(p.birth_rate, (pd.Series, pd.DataFrame)):
            available_years = p.birth_rate.index
            year_ind = sc.findnearest(available_years, sim.t.now('year'))
            nearest_year = available_years[year_ind]
            this_birth_rate = p.birth_rate.loc[nearest_year]
            scaled_birth_prob = ss.prob.array_to_prob(this_birth_rate, self.t.dt, v=p.rate_units * p.rel_birth)
        else:
            this_birth_rate = p.birth_rate
            if isinstance(this_birth_rate, ss.Rate):
                this_birth_rate = ss.prob(this_birth_rate.value * self.pars.rate_units * self.pars.rel_birth, this_birth_rate.unit).to_prob(ss.years(1))
            else: # number
                this_birth_rate = this_birth_rate * self.pars.rate_units * self.pars.rel_birth

            scaled_birth_prob = ss.prob.array_to_prob(np.array([this_birth_rate]), self.t.dt)[0]

        scaled_birth_prob = np.clip(scaled_birth_prob, a_min=0, a_max=1)
        self.dist.set(p=scaled_birth_prob) # Update the distribution with the probabilities for this timestep
        birth_uids = self.dist.filter()
        # n_new = np.random.binomial(n=sim.people.alive.count(), p=scaled_birth_prob) # Not CRN safe, see issue #404
        return birth_uids

    def step(self):
        new_uids = self.add_births()
        self.n_births = len(new_uids)
        return new_uids

    def add_births(self):
        """ Add n_new births to each state in the sim """
        people = self.sim.people
        birth_uids = self.get_births()
        new_uids = people.grow(len(birth_uids))
        people.age[new_uids] = 0
        people.parent[new_uids] = birth_uids
        return new_uids

    def update_results(self):
        """ Calculate new births and crude birth rate """
        # New births -- already calculated
        self.results.new[self.ti] = self.n_births

        # Calculate crude birth rate (CBR)
        inv_rate_units = 1.0/self.pars.rate_units
        births_per_year = self.n_births/self.sim.t.dt_year
        denom = self.sim.people.alive.sum()
        self.results.cbr[self.ti] = inv_rate_units*births_per_year/denom
        return

    def finalize(self):
        super().finalize()
        self.results.cumulative[:] = np.cumsum(self.results.new)
        return


class Deaths(Demographics):
    """
    Configure disease-independent "background" deaths.

    The probability of death for each agent on each timestep is determined
    by the `death_rate` parameter and the time step. The default value of
    this parameter is 0.01, indicating that all agents will
    face a 1% chance of death per year.

    However, this function can be made more realistic by using a dataframe
    for the `death_rate` parameter, to allow it to vary by year, sex, and
    age.  The separate 'metadata' argument can be used to configure the
    details of the input datafile.

    Alternatively, it is possible to override the `death_rate` parameter
    with a bernoulli distribution containing a constant value of function of
    your own design.

    Args:
        pars: dict with arguments including:
            rel_death: constant used to scale all death rates
            death_rate: float, dict, or pandas dataframe/series containing mortality data
            rate_units: units for death rates (see in-line comment on par dict below)

        metadata: data about the data contained within the data input.
            "data_cols" is is a dictionary mapping standard keys, like "year" to the
            corresponding column name in data. Similar for "sex_keys". 
    """
    def __init__(self, pars=None, rel_death=_, death_rate=_, rate_units=_, metadata=None, **kwargs):
        super().__init__()
        self.define_pars(
            rel_death = 1,
            death_rate = ss.peryear(10),  # Default = a fixed rate of 2%/year, overwritten if data provided
            rate_units = 1e-3,  # assumes death rates are per 1000. If using percentages, switch this to 1
        )
        self.update_pars(pars, **kwargs)

        # Process metadata. Defaults here are the labels used by UN data
        self.metadata = sc.mergedicts(
            sc.objdict(
                data_cols = dict(year='Time', sex='Sex', age='AgeGrpStart', value='mx'),
                sex_keys = {'Female':'f', 'Male':'m'},
            ),
            metadata
        )

        # Process data, which may be provided as a number, dict, dataframe, or series
        # If it's a number it's left as-is; otherwise it's converted to a dataframe
        self.death_rate_data = self.standardize_death_data() # TODO: refactor
        self.pars.p_death = ss.bernoulli(p=0)  # Placeholder, populated by make_p_death
        self.n_deaths = 0 # For results tracking
        return

    def standardize_death_data(self):
        """ Standardize/validate death rates - handled in an external file due to shared functionality """
        death_rate = ss.standardize_data(data=self.pars.death_rate, metadata=self.metadata)
        if isinstance(death_rate, (pd.Series, pd.DataFrame)):
            death_rate = death_rate.unstack(level='age')
            assert not death_rate.isna().any(axis=None) # For efficiency, we assume that the age bins are the same for all years in the input dataset
        if sc.isnumber(death_rate):
            # If the user has provided a bare number, assume it is per year
            msg = f'Death rate was specified as a number rather than a rate - assuming it is {death_rate} per year'
            ss.warn(msg)
            return ss.peryear(death_rate)
        return death_rate

    def make_p_death(self):
        """ Take in the module, sim, and uids, and return the probability of death for each UID on this timestep """
        sim = self.sim
        drd = self.death_rate_data

        # We are going to set up death_rate as a prob with unit=ss.years(1). That is the probability of death in 1 year
        if sc.isnumber(drd):
            death_rate = np.array([drd * self.pars.rate_units * self.pars.rel_death])  # Later gets interpreted as a prob with unit=ss.years(1) by ss.prob.array_to_prob
        elif isinstance(drd, ss.Rate):
            if drd.unit == 1: # drd is already in years so don't need to translate
                death_rate = np.array([drd.value * self.pars.rate_units * self.pars.rel_death])
            else:
                # Convert from prob per drd.unit to prob per year, ss.prob.array_to_prob will convert to prob per timestep later
                death_rate = np.array([ss.prob(drd.value * self.pars.rate_units * self.pars.rel_death, drd.unit).to_prob(ss.years(1))])

        # Process data
        else:
            ppl = sim.people
            uids = ppl.auids  # Get the UIDs of all alive people

            available_years = drd.index.get_level_values('year')
            year_ind = sc.findnearest(available_years, sim.t.now('year'))  # TODO: make work with different timesteps
            nearest_year = available_years[year_ind]

            death_rate = np.empty(uids.shape, dtype=ss_float)

            if 'sex' in drd.index.names:
                s = drd.loc[nearest_year, 'f']
                binned_ages = np.digitize(ppl.age[ppl.female], s.index)-1 # Negative ages will be in the first bin - do *not* subtract 1 so that this bin is 0
                death_rate[ppl.female] = s.values[binned_ages]
                s = drd.loc[nearest_year, 'm']
                binned_ages = np.digitize(ppl.age[ppl.male], s.index)-1 # Negative ages will be in the first bin - do *not* subtract 1 so that this bin is 0
                death_rate[ppl.male] = s.values[binned_ages]
            else:
                s = drd.loc[nearest_year]
                binned_ages = np.digitize(ppl.age, s.index)-1 # Negative ages will be in the first bin - do *not* subtract 1 so that this bin is 0
                death_rate[:] = s.values[binned_ages]
            death_rate *= self.pars.rate_units * self.pars.rel_death

        # Scale from rate to probability
        death_rate = ss.peryear(death_rate)
        p_death = death_rate.to_prob(self.t.dt)  # Convert to probability per timestep

        if sc.isnumber(drd) or isinstance(drd, ss.Rate):
            p_death = p_death[0] # TODO: what???
        return p_death

    def init_results(self):
        super().init_results()
        self.define_results(
            ss.Result('new',        dtype=int,   scale=True,  summarize_by='sum',  label='Deaths', auto_plot=False), # Use sim deaths instead
            ss.Result('cumulative', dtype=int,   scale=True,  summarize_by='last', label='Cumulative deaths', auto_plot=False),
            ss.Result('cmr',        dtype=float, scale=False, summarize_by='mean', label='Crude mortality rate'),
        )
        return

    def step(self):
        """ Select people to die """
        p_death = self.make_p_death()  # Get the probability of death for each agent
        self.pars.p_death.set(p=p_death)  # Update the distribution with the probabilities for this timestep
        death_uids = self.pars.p_death.filter()
        self.sim.people.request_death(death_uids)
        self.n_deaths = len(death_uids)
        return self.n_deaths

    def update_results(self):
        self.results['new'][self.ti] = self.n_deaths
        return

    def finalize(self):
        super().finalize()
        self.results.cumulative[:] = np.cumsum(self.results.new)
        units = self.pars.rate_units*self.sim.t.dt_year
        inds = self.match_time_inds()
        n_alive = self.sim.results.n_alive[inds]
        deaths = np.divide(self.results.new, n_alive, where=n_alive>0)
        self.results.cmr[:] = deaths/units
        return


class Pregnancy(Demographics):
    """
    Create births via pregnancies for each agent.

    This module models conception, pregnancy, and birth at the individual level using
    age-specific fertility rates. Supports prenatal and postnatal transmission
    networks, maternal and neonatal mortality, and burn-in of existing
    pregnancies at simulation start.

    Args:
        dur_pregnancy (float/dur): duration of pregnancy (default 9 months)
        dur_postpartum (float/dur): duration of postpartum period for postnatal transmission (default 6 months)
        fertility_rate (float/dataframe): value or dataframe with age-specific fertility rates
        rel_fertility (float): constant used to scale all fertility rates
        p_maternal_death (float): probability of maternal death during pregnancy (default 0.0)
        p_neonatal_death (float): probability of neonatal death (default 0.0)
        sex_ratio (float): probability of female births (default 0.5)
        min_age (float): minimum age for pregnancy (default 15)
        max_age (float): maximum age for pregnancy (default 50)
        rate_units (float): units for fertility rates (default assumes per 1000)
        burnin (bool): whether to seed pregnancies from before simulation start (default true)
        metadata (dict): data column mappings for fertility rate data if a dataframe is supplied
    """
    def __init__(self, pars=None, dur_pregnancy=_, dur_postpartum=_, fertility_rate=_, rel_fertility=_,
                 p_maternal_death=_, p_neonatal_death=_, sex_ratio=_, min_age=_, max_age=_,
                 rate_units=_, burnin=_, slot_scale=_, min_slots=_, metadata=None, **kwargs):
        super().__init__()
        self.define_pars(
            dur_pregnancy = ss.years(0.75), # Duration for pre-natal transmission
            dur_postpartum = ss.lognorm_ex(mean=ss.years(0.5), std=ss.years(0.5)), # Duration for post-natal transmission (e.g. via breastfeeding)
            fertility_rate = ss.peryear(100), # Can be a number of Pandas DataFrame
            rel_fertility = 1,
            p_maternal_death = ss.bernoulli(0),
            p_neonatal_death = ss.bernoulli(0),
            sex_ratio = ss.bernoulli(0.5), # Ratio of babies born female
            min_age = 15, # Minimum age to become pregnant
            max_age = 50, # Maximum age to become pregnant
            rate_units = 1e-3, # Assumes fertility rates are per 1000. If using percentages, switch this to 1
            burnin = True, # Should we seed pregnancies that would have happened before the start of the simulation?
            slot_scale = 5, # Random slots will be assigned to newborn agents between min=n_agents and max=slot_scale*n_agents
            min_slots  = 100, # Minimum number of slots, useful if the population size is very small
        )
        self.update_pars(pars, **kwargs)

        self.pars.p_fertility = ss.bernoulli(p=0) # Placeholder, see make_p_fertility

        # Other, e.g. postpartum, on contraception...
        self.define_states(
            ss.BoolState('fecund', default=True, label='Female of childbearing age'),
            ss.BoolState('pregnant', label='Pregnant'),  # Currently pregnant
            ss.BoolState('postpartum', label="Post-partum"),  # Currently post-partum
            ss.FloatArr('parity', label='Parity', default=0),  # Number of pregnancies
            ss.FloatArr('child_uid', label='UID of children, from embryo through postpartum'),
            ss.FloatArr('dur_postpartum', label='Post-partum duration'),  # Duration of postpartum phase
            ss.FloatArr('ti_pregnant', label='Time of pregnancy'),  # Time pregnancy begins
            ss.FloatArr('ti_delivery', label='Time of delivery'),  # Time of delivery
            ss.FloatArr('ti_postpartum', label='Time post-partum ends'),  # Time postpartum ends
            ss.FloatArr('ti_dead', label='Time of maternal death'),  # Maternal mortality
        )

        # Process metadata. Defaults here are the labels used by UN data
        self.metadata = sc.mergedicts(
            sc.objdict(data_cols=dict(year='Time', age='AgeGrp', value='ASFR')),
            metadata,
        )
        self.choose_slots = None # Distribution for choosing slots; set in self.init()

        # For results tracking
        self.n_pregnancies = 0
        self.n_births = 0
        return

    def make_p_fertility(self, filter_uids=None):
        """ Take in the module, sim, and uids, and return the conception probability for each UID on this timestep """
        sim = self.sim
        ppl = sim.people

        # Filter UIDS and age
        may_conceive = ppl.female & (ppl.age >= self.pars.min_age) & (ppl.age <= self.pars.max_age)
        uids = may_conceive
        if filter_uids is not None: uids = filter_uids & uids
        age = sim.people.age[uids]

        # Get data, check it's in the right form
        frd = self.fertility_rate_data
        if sc.isnumber(frd):
            raise TypeError('Fertility rate should be specified as a rate (or with time-varying data)')

        # Initialize rate as an array the length of raw UIDs, so we can index it with UIDs
        fertility_rate = np.zeros(len(sim.people.uid.raw), dtype=ss_float)

        if isinstance(frd, ss.Rate):
            fertility_rate[uids] = ss.prob(frd.value * (self.pars.rate_units * self.pars.rel_fertility), frd.unit).to_prob(ss.years(1))
        else:
            conception_time = self.t.year-self.pars.dur_pregnancy.years
            if isinstance(conception_time, ss.years):
                conception_time = conception_time.years
            year_ind = sc.findnearest(frd.index, conception_time)
            nearest_year = frd.index[year_ind]

            # Assign agents to age bins
            age_bins = self.fertility_rate_data.columns.values
            age_bin_all = np.digitize(age, age_bins) - 1
            new_rate = self.fertility_rate_data.loc[nearest_year].values.copy()  # Initialize array with new rates

            if (~self.fecund).any():
                # Scale the new rate to convert the denominator from all women to fecund women
                v, c = np.unique(age_bin_all, return_counts=True)
                age_counts = np.zeros(len(age_bins))
                age_counts[v] = c

                age_bin_infecund = np.digitize(sim.people.age[~self.fecund], age_bins) - 1
                v, c = np.unique(age_bin_infecund, return_counts=True)
                infecund_age_counts = np.zeros(len(age_bins))
                infecund_age_counts[v] = c

                num_to_make = new_rate * age_counts  # Number that we need to make pregnant
                new_denom = age_counts - infecund_age_counts  # New denominator for rates
                np.divide(num_to_make, new_denom, where=new_denom>0, out=new_rate)

            fertility_rate[uids] = new_rate[age_bin_all] * (self.pars.rate_units * self.pars.rel_fertility)  # Prob per year

        # Scale from rate to probability
        fertility_rate[(~self.fecund).uids] = 0  # Currently infecund women cannot become pregnant
        fertility_rate = ss.peryear(fertility_rate[filter_uids])  # Only return rates for requested UIDs
        p_fertility = fertility_rate.to_prob(self.t.dt)  # Convert to probability per timestep
        return p_fertility

    def standardize_fertility_data(self):
        """
        Standardize/validate fertility rates
        """
        fertility_rate = ss.standardize_data(data=self.pars.fertility_rate, metadata=self.metadata)
        if isinstance(fertility_rate, (pd.Series, pd.DataFrame)):
            fertility_rate = fertility_rate.unstack()
            # Interpolate to 1 year increments
            fertility_rate = fertility_rate.reindex(np.arange(fertility_rate.index.min(), fertility_rate.index.max() + 1)).interpolate()
            max_age = fertility_rate.columns.max()
            fertility_rate[max_age + 1] = 0
            assert not fertility_rate.isna().any(axis=None) # For efficiency, we assume that the age bins are the same for all years in the input dataset
        if sc.isnumber(fertility_rate):
            msg = f'Fertility rate was specified as a number rather than a rate - assuming it is {fertility_rate} per year'
            ss.warn(msg)
            return ss.peryear(fertility_rate)
        return fertility_rate

    def init_pre(self, sim):
        super().init_pre(sim)

        # Process data, which may be provided as a number, dict, dataframe, or series
        # If it's a number it's left as-is; otherwise it's converted to a dataframe
        self.fertility_rate_data = self.standardize_fertility_data()

        low = sim.pars.n_agents + 1
        high = int(self.pars.slot_scale*sim.pars.n_agents)
        high = np.maximum(high, self.pars.min_slots) # Make sure there are at least min_slots slots to avoid artifacts related to small populations
        self.choose_slots = ss.randint(low=low, high=high, sim=sim, module=self)
        return

    def init_results(self):
        """
        Results could include a range of birth outcomes e.g. LGA, stillbirths, etc.
        Still unclear whether this logic should live in the pregnancy module, the
        individual disease modules, the connectors, or the sim.
        """
        super().init_results()
        self.define_results(
            ss.Result('pregnancies', dtype=int,   scale=True,  summarize_by='sum',  label='New pregnancies'),
            ss.Result('births',      dtype=int,   scale=True,  summarize_by='sum',  label='New births'),
            ss.Result('cbr',         dtype=float, scale=False, summarize_by='mean', label='Crude birth rate'),
        )
        return

    def step(self):
        if self.ti == 0 and self.pars.burnin: # TODO: refactor
            dtis = np.arange(np.ceil(-1 * self.pars.dur_pregnancy/self.dt), 0, 1).astype(int)
            for dti in dtis:
                self.t.ti = dti
                self.do_step()
            self.t.ti = 0
        new_uids = self.do_step()
        return new_uids

    def do_step(self):
        """ Perform all updates """
        self.update_states()
        conceive_uids = self.make_pregnancies()
        self.n_pregnancies += len(conceive_uids) # += to handle burn-in
        new_uids = self.make_embryos(conceive_uids)
        return new_uids

    def update_states(self):
        """ Update states """
        # Check for new deliveries
        ti = self.ti
        deliveries = self.pregnant & (self.ti_delivery <= ti)
        self.n_births = np.count_nonzero(deliveries)
        self.pregnant[deliveries] = False
        self.postpartum[deliveries] = True
        self.fecund[deliveries] = False
        self.parity[deliveries] += 1  # Increment parity for the mothers

        # Add connections to any postnatal transmission layers
        for lkey, layer in self.sim.networks.items():
            if layer.postnatal and self.n_births:

                # Add postnatal connections by finding the prenatal contacts
                # Validation of the networks is done during initialization to ensure that 1 prenatal netwrok is present
                prenatalnet = [nw for nw in self.sim.networks.values() if nw.prenatal][0]

                # Find the prenatal connections that are ending
                prenatal_ending = prenatalnet.edges.end <= ti
                new_mother_uids = prenatalnet.edges.p1[prenatal_ending]
                new_infant_uids = prenatalnet.edges.p2[prenatal_ending]

                # Validation
                if not set(new_mother_uids) == set(deliveries.uids): # Not sure why sometimes out of order
                    errormsg = 'IDs of new mothers do not match IDs of new deliveries'
                    raise ValueError(errormsg)

                # Create durations and start dates, and add connections
                durs = self.dur_postpartum[new_mother_uids]
                start = np.full(self.n_births, fill_value=ti)

                # # Remove pairs from prenatal network and add to postnatal
                prenatalnet.end_pairs()
                layer.add_pairs(new_mother_uids, new_infant_uids, dur=durs, start=start)

        # Check for new women emerging from post-partum
        postpartum = self.postpartum & (self.ti_postpartum <= ti)
        self.postpartum[postpartum] = False
        self.fecund[postpartum] = True
        self.child_uid[postpartum] = np.nan

        # Maternal deaths
        maternal_deaths = (self.ti_dead <= ti).uids
        self.sim.people.request_death(maternal_deaths)
        return

    def make_pregnancies(self):
        """ Select people to make pregnant using incidence data """
        # People eligible to become pregnant. We don't remove pregnant people here, these
        # are instead handled in the fertility_dist logic as the rates need to be adjusted
        eligible_uids = self.sim.people.female.uids
        p_fertility = self.make_p_fertility(eligible_uids)
        self.pars.p_fertility.set(p_fertility)
        dist = self.pars.p_fertility
        conceive_uids = dist.filter(eligible_uids)

        if len(conceive_uids) == 0:
            return ss.uids()

        # Validation
        if np.any(self.pregnant[conceive_uids]):
            which_uids = conceive_uids[self.pregnant[conceive_uids]]
            errormsg = f'New conceptions registered in {len(which_uids)} pregnant agent(s) at timestep {self.ti}.'
            raise ValueError(errormsg)

        # Set prognoses for the pregnancies
        self.set_prognoses(conceive_uids)
        return conceive_uids

    def make_embryos(self, conceive_uids):
        """ Add properties for the just-conceived """
        people = self.sim.people
        n_unborn = len(conceive_uids)
        if n_unborn == 0:
            new_uids = ss.uids()
        else:

            # Choose slots for the unborn agents
            new_slots = self.choose_slots.rvs(conceive_uids)

            # Grow the arrays and set properties for the unborn agents
            new_uids = people.grow(len(new_slots), new_slots)
            people.age[new_uids] = -self.pars.dur_pregnancy.years
            people.slot[new_uids] = new_slots  # Before sampling female_dist
            people.female[new_uids] = self.pars.sex_ratio.rvs(conceive_uids)
            people.parent[new_uids] = conceive_uids
            self.child_uid[conceive_uids] = new_uids

            # Add connections to any prenatal transmission layers
            for lkey, layer in self.sim.networks.items():
                if layer.prenatal:
                    durs = np.full(n_unborn, fill_value=self.pars.dur_pregnancy/layer.t.dt) # Network edge duration is stored in units of timesteps, so we need to convert dur_pregnancy to a number of timesteps here
                    start = np.full(n_unborn, fill_value=self.ti)
                    layer.add_pairs(conceive_uids, new_uids, dur=durs, start=start)

        if self.ti < 0:
            people.age[new_uids] += -self.ti * self.sim.t.dt_year # Age to ti=0

        return new_uids

    def set_prognoses(self, uids):
        """
        Make pregnancies
        Add miscarriage/termination logic here
        Also reconciliation with birth rates
        Q, is this also a good place to check for other conditions and set prognoses for the fetus?
        """

        # Change states for the newly pregnant woman
        ti = self.ti
        self.fecund[uids] = False
        self.pregnant[uids] = True
        self.ti_pregnant[uids] = ti

        # Outcomes for pregnancies
        dur_preg = self.pars.dur_pregnancy
        dur_postpartum = self.pars.dur_postpartum.rvs(uids)
        dead = self.pars.p_maternal_death.rvs(uids)
        self.ti_delivery[uids] = ti + dur_preg/self.t.dt # Currently assumes maternal deaths still result in a live baby. Note that since we are adding the duration to `ti` (the number of timesteps) we need to convert `dur_preg` into a number of timesteps
        self.ti_postpartum[uids] = self.ti_delivery[uids] + dur_postpartum
        self.dur_postpartum[uids] = dur_postpartum

        if np.any(dead): # NB: 100x faster than np.sum(), 10x faster than np.count_nonzero()
            self.ti_dead[uids[dead]] = ti + dur_preg
        return

    def finish_step(self):
        super().finish_step()
        death_uids = ss.uids(self.sim.people.ti_dead <= self.ti)
        if len(death_uids) == 0:
            return

        # Any pregnant? Consider death of the neonate
        mother_death_uids = death_uids[self.pregnant[death_uids]]
        if len(mother_death_uids):
            neonate_uids = ss.uids(self.child_uid[mother_death_uids])
            neonatal_death_uids = self.pars.p_neonatal_death.filter(neonate_uids)
            if len(neonatal_death_uids):
                self.sim.people.request_death(neonatal_death_uids)

        # Any prenatal? Handle changes to pregnancy
        is_prenatal = self.sim.people.age[death_uids] < 0
        prenatal_death_uids = death_uids[is_prenatal]
        if len(prenatal_death_uids):
            mother_uids = self.sim.people.parent[prenatal_death_uids]
            self.pregnant[mother_uids] = False # Baby lost, mother no longer pregnant
            self.fecund[mother_uids] = True # Or wait?
            self.postpartum[mother_uids] = False
            self.child_uid[mother_uids] = np.nan
            self.ti_delivery[mother_uids] = np.nan
            self.ti_postpartum[mother_uids] = np.nan
        return

    def update_results(self):
        ti = self.ti
        self.results['pregnancies'][ti] = self.n_pregnancies
        self.results['births'][ti] = self.n_births

        # Reset for the next step
        self.n_pregnancies = 0
        self.n_births = 0

        return

    def finalize(self):
        super().finalize()
        units = self.pars.rate_units*self.sim.t.dt_year
        inds = self.match_time_inds()
        n_alive = self.sim.results.n_alive[inds]
        births = np.divide(self.results['births'], n_alive, where=n_alive>0)
        self.results['cbr'][:] = births/units
        return



import starsim as ss

pars = dict(
    diseases = ss.SIS(unit='day', dt=1.0, init_prev=0.1),
    demographics = ss.Births(dt=0.25, unit='year'),
    networks = ss.RandomNet(unit='week'),
)

sim = ss.Sim(pars, unit='day', dt=2, start='2000-01-01', stop='2002-01-01')
sim.run()


"""
Numerical utilities and other helper functions
"""
import warnings
import numpy as np
import pandas as pd
import sciris as sc
import matplotlib as mpl
import matplotlib.pyplot as plt
import starsim as ss

# %% Helper functions

# What functions are externally visible
__all__ = ['ndict', 'warn', 'find_contacts', 'standardize_netkey', 'standardize_data',
           'validate_sim_data', 'load', 'save', 'plot_args', 'show', 'return_fig']

class ndict(sc.objdict):
    """
    A dictionary-like class that provides additional functionalities for handling named items.

    Args:
        name (str): The attribute of the item to use as the dict key (i.e., all items should have this attribute defined)
        type (type): The expected type of items.
        strict (bool): If True, only items with the specified attribute will be accepted.
        overwrite (bool): whether to allow adding a key when one has already been added

    **Examples**:

        networks = ss.ndict(ss.MFNet(), ss.MaternalNet())
        networks = ss.ndict([ss.MFNet(), ss.MaternalNet()])
        networks = ss.ndict({'mf':ss.MFNet(), 'maternal':ss.MaternalNet()})
    """

    def __init__(self, *args, nameattr='name', type=None, strict=True, overwrite=False, **kwargs):
        super().__init__()
        self.setattribute('_nameattr', nameattr)  # Since otherwise treated as keys
        self.setattribute('_type', type)
        self.setattribute('_strict', strict)
        self.setattribute('_overwrite', overwrite)
        self.extend(*args, **kwargs)
        return

    def __call__(self):
        """ Shortcut for returning values """
        return self.values()

    def append(self, arg, key=None, overwrite=None):
        valid = False
        if arg is None:
            return  # Nothing to do
        elif hasattr(arg, self._nameattr):
            key = key or getattr(arg, self._nameattr)
            valid = True
        elif isinstance(arg, dict):
            if self._nameattr in arg:
                key = key or arg[self._nameattr]
                valid = True
            else:
                for k, v in arg.items():
                    self.append(v, key=k)
                valid = None  # Skip final processing
        elif not self._strict:
            key = key or f'item{len(self) + 1}'
            valid = True
        else:
            valid = False

        if valid is True:
            if self._strict:
                self._check_type(arg)
                self._check_key(key, overwrite=overwrite)
            self[key] = arg # Actually add to the ndict!
        elif valid is None:
            pass  # Nothing to do
        else:
            errormsg = f'Could not interpret argument {arg}: does not have expected attribute "{self._nameattr}"'
            raise TypeError(errormsg)
        return self

    def _check_key(self, key, overwrite=None):
        if overwrite is None: overwrite = self._overwrite
        if key in self:
            if not overwrite:
                typestr = f' "{self._type}"' if self._type else ''
                errormsg = f'Cannot add object "{key}" since already present in ndict{typestr} with keys:\n{sc.newlinejoin(self.keys())}'
                raise ValueError(errormsg)
            else:
                ss.warn(f'Overwriting existing ndict entry "{key}"')
        return

    def _check_type(self, arg):
        """ Check types """
        if self._type is not None:
            if not isinstance(arg, self._type) and not isinstance(arg, ss.Module): # Module is a valid argument anywhere
                errormsg = f'The following item does not have the expected type {self._type}:\n{arg}'
                raise TypeError(errormsg)
        return

    def extend(self, *args, **kwargs):
        """ Add new items to the ndict, by item, list, or dict """
        args = sc.mergelists(*args)
        for arg in args:
            self.append(arg)
        for key, arg in kwargs.items():
            self.append(arg, key=key)
        return self

    def merge(self, other):
        """ Merge another dictionary with this one """
        for key, arg in other.items():
            self.append(arg, key=key)
        return self

    def copy(self):
        """ Shallow copy """
        new = self.__class__.__new__(nameattr=self._nameattr, type=self._type, strict=self._strict)
        new.update(self)
        return new

    def __add__(self, other):
        """ Allow c = a + b """
        new = self.copy()
        if isinstance(other, list):
            new.extend(other)
        else:
            new.append(other)
        return new

    def __iadd__(self, other):
        """ Allow a += b """
        if isinstance(other, list):
            self.extend(other)
        else:
            self.append(other)
        return self


def nlist_to_dict(nlist, die=True):
    """ Convert a list of named items (e.g. modules, states) to a dictionary; not for the user """
    out = sc.objdict()
    collisions = []
    for obj in nlist:
        key = obj.name
        if key not in out:
            out[key] = obj
        else:
            collisions.append(key)
    if collisions:
        errormsg = f'Cannot list modules as a dict since one or more modules have the same name; use the list version instead. Collisions:\n{collisions}'
        if die:
            raise ValueError(errormsg)
        else:
            ss.warn(errormsg)
    else:
        return out


def warn(msg, category=None, verbose=None, die=None):
    """ Helper function to handle warnings -- shortcut to warnings.warn """

    # Handle inputs
    warnopt = ss.options.warnings if not die else 'error'
    if category is None:
        category = RuntimeWarning
    if verbose is None:
        verbose = ss.options.verbose

    # Handle the different options
    if warnopt in ['error', 'errors']:  # Include alias since hard to remember
        raise category(msg)
    elif warnopt == 'warn':
        msg = '\n' + msg
        warnings.warn(msg, category=category, stacklevel=2)
    elif warnopt == 'print':
        if verbose:
            msg = 'Warning: ' + msg
            print(msg)
    elif warnopt == 'ignore':
        pass
    else:
        options = ['error', 'warn', 'print', 'ignore']
        errormsg = f'Could not understand "{warnopt}": should be one of {options}'
        raise ValueError(errormsg)

    return


def find_contacts(p1, p2, inds):  # pragma: no cover
    """
    Variation on Network.find_contacts() that avoids sorting.

    A set is returned here rather than a sorted array so that custom tracing interventions can efficiently
    add extra people. For a version with sorting by default, see Network.find_contacts(). Indices must be
    an int64 array since this is what's returned by true() etc. functions by default.
    """
    pairing_partners = set()
    inds = set(inds)
    for i in range(len(p1)):
        if p1[i] in inds:
            pairing_partners.add(p2[i])
        if p2[i] in inds:
            pairing_partners.add(p1[i])
    return pairing_partners


# %% Data cleaning and processing

def standardize_netkey(key):
    """ Networks can be upper or lowercase, and have a suffix 'net' or not; this function standardizes them """
    return key.lower().removesuffix('net')


def standardize_data(data=None, metadata=None, min_year=1800, out_of_range=0, default_age=0, default_year=2024):
    """
    Standardize formats of input data

    Input data can arrive in many different forms. This function accepts a variety of data
    structures, and converts them into a Pandas Series containing one variable, based on
    specified metadata, or an `ss.Dist` if the data is already an `ss.Dist` object.

    The metadata is a dictionary that defines columns of the dataframe or keys
    of the dictionary to use as indices in the output Series. It should contain:

    - `metadata['data_cols']['value']` specifying the name of the column/key to draw values from
    - `metadata['data_cols']['year']` optionally specifying the column containing year values; otherwise the default year will be used
    - `metadata['data_cols']['age']` optionally specifying the column containing age values; otherwise the default age will be used
    - `metadata['data_cols'][<arbitrary>]` optionally specifying any other columns to use as indices. These will form part of the multi-index for the standardized Series output.

    If a `sex` column is part of the index, the metadata can also optionally specify a string mapping to convert
    the sex labels in the input data into the 'm'/'f' labels used by Starsim. In that case, the metadata can contain
    an additional key like `metadata['sex_keys'] = {'Female':'f','Male':'m'}` which in this case would map the strings
    'Female' and 'Male' in the original data into 'm'/'f' for Starsim.

    Args:
        data (pandas.DataFrame, pandas.Series, dict, int, float): An associative array  or a number, with the input data to be standardized.
        metadata (dict): Dictionary specifiying index columns, the value column, and optionally mapping for sex labels
        min_year (float): Optionally specify a minimum year allowed in the data. Default is 1800.
        out_of_range (float): Value to use for negative ages - typically 0 is a reasonable choice but other values (e.g., np.inf or np.nan) may be useful depending on the calculation. This will automatically be added to the dataframe with an age of `-np.inf`

    Returns:

        - A `pd.Series` for all supported formats of `data` *except* an `ss.Dist`. This series will contain index columns for 'year'
          and 'age' (in that order) and then subsequent index columns for any other variables specified in the metadata, in the order
          they appeared in the metadata (except for year and age appearing first).
        - An `ss.Dist` instance - if the `data` input is an `ss.Dist`, that same object will be returned by this function
    """
    # It's a format that can be used directly: return immediately
    if sc.isnumber(data) or isinstance(data, (ss.Dist, ss.Rate)):
        return data

    # Convert series and dataframe inputs into dicts
    if isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
        data = data.reset_index().to_dict(orient='list')

    # Check that the input is now a dict (scalar types have already been handled above
    assert isinstance(data, dict), 'Supported inputs are ss.Dict, scalar numbers, DataFrames, Series, or dictionaries'

    # Extract the values and index columns
    assert 'value' in metadata['data_cols'], 'The metadata is missing a column name for "value", which must be provided if the input data is in the form of a DataFrame, Series, or dict'
    values = sc.promotetoarray(data[metadata['data_cols']['value']])
    index = sc.objdict()
    for k, col in metadata['data_cols'].items():
        if k != 'value':
            index[k] = data[col]

    # Add defaults for year and age
    if 'year' not in index:
        index['year'] = np.full(values.shape, fill_value=default_year, dtype=ss.dtypes.float)
    if 'age' not in index:
        index['age'] = np.full(values.shape, fill_value=default_age, dtype=ss.dtypes.float)

    # Reorder the index so that it starts with age first
    index.insert(0, 'age', index.pop('age'))
    index.insert(0, 'year', index.pop('year'))

    # Map sex values
    if 'sex' in index:
        if 'sex_keys' in metadata:
            index['sex'] = [metadata['sex_keys'][x] for x in index['sex']]
        assert set(index['sex']) == {'f','m'}, 'If standardized data contains a "sex" column, it should use "m" and "f" to specify the sex.'

    # Create the series
    output = pd.Series(data=values, index=pd.MultiIndex.from_arrays(index.values(), names=index.keys()))

    # Add an entry for negative ages
    new_entries = output.index.set_codes([0] * len(output), level=1).set_levels([-np.inf], level=1).unique()
    new = pd.Series(data=out_of_range, index=new_entries)
    output = pd.concat([output, new])

    # Truncate years
    output = output.loc[output.index.get_level_values('year')>min_year]

    # Perform a final sort to prevent "indexing past lexsort depth" warnings
    output = output.sort_index()

    return output


def validate_sim_data(data=None, die=None):
    """
    Validate data intended to be compared to the sim outputs, e.g. for calibration

    Args:
        data (df/dict): a dataframe (or dict) of data, with a column "time" plus data columns of the form "module.result", e.g. "hiv.new_infections"
        die (bool): whether to raise an exception if the data cannot be converted (default: die if data is not None but cannot be converted)

    """
    success = False
    if data is not None:
        # Try loading the data
        try:
            data = sc.dataframe(data) # Convert it to a dataframe
            timecols = ['t', 'timevec', 'tvec', 'time', 'day', 'date', 'year'] # If a time column is supplied, use it as the index
            found = False
            for timecol in timecols:
                if timecol in data.cols:
                    if found:
                        errormsg = f'Multiple time columns found: please ensure only one of {timecols} is present; you supplied {data.cols}.'
                        raise ValueError(errormsg)
                    data.set_index(timecol, inplace=True)
                    found = True
            success = True

        # Data loading failed
        except Exception as E:
            errormsg = f'Failed to add data "{data}": expecting a dataframe-compatible object. Error:\n{E}'
            if die == False:
                print(errormsg)
            else:
                raise ValueError(errormsg)

    # Validation
    if not success and die == True:
        errormsg = 'Data "{data}" could not be converted and die == True'
        raise ValueError(errormsg)

    return data


def combine_rands(a, b):
    """
    Efficient algorithm for combining two arrays of random integers into an array
    of floats.

    See ss.multi_random() for the user-facing version.

    Args:
        a (array): array of random integers between 0 and np.iinfo(np.uint64).max, as from ss.rand_raw()
        b (array): ditto, same size as a

    Returns:
        A new array of random numbers the same size as a and b
    """
    c = np.bitwise_xor(a*b, a-b)
    u = c / np.iinfo(np.uint64).max
    return u


#%% Other helper functions

def load(filename, **kwargs):
    """
    Alias to Sciris `sc.loadany()`

    Since Starsim uses Sciris for saving objects, they can be loaded back using
    this function. This can also be used to load other objects of known type
    (e.g. JSON), although this usage is discouraged.

    Args:
        filename (str/path): the name of the file to load
        kwargs (dict): passed to `sc.loadany()`

    Returns:
        The loaded object
    """
    return sc.loadany(filename, **kwargs)


def save(filename, obj, **kwargs):
    """
    Alias to Sciris `sc.save()`

    While some Starsim objects have their own save methods, this function can be
    used to save any arbitrary object. It can then be loaded with ss.load().

    Args:
        filename (str/path): the name of the file to save
        obj (any): the object to save
        kwargs (dict): passed to `sc.save()`
    """
    return sc.save(filename=filename, obj=obj, **kwargs)


class shrink:
    """ Define a class to indicate an object has been shrunken """
    def __repr__(self):
        s = 'This object has been intentionally "shrunken"; it is a placeholder and has no functionality. Use the non-shrunken object instead.'
        return s


#%% Plotting helper functions

# Specify known/common keywords
plotting_kw = sc.objdict()
plotting_kw.fig = ['figsize', 'nrows', 'ncols', 'ratio', 'num', 'dpi', 'facecolor'] # For sc.getrowscols()
plotting_kw.plot = ['alpha', 'c', 'lw', 'linewidth', 'marker', 'markersize', 'ms'] # For plt.plot()
plotting_kw.data = {'data_alpha':'alpha', 'data_color':'color', 'data_size':'markersize'} # For plt.scatter()
plotting_kw.fill = {'fill_alpha':'alpha', 'fill_color':'color', 'fill_hatch':'hatch', 'fill_lw':'lw'}
plotting_kw.legend = ['loc', 'bbox_to_anchor', 'ncols', 'reverse', 'frameon']
plotting_kw.style = ['style', 'font', 'fontsize', 'interactive'] # For sc.options.with_style()
plotting_kw.return_fig = ['do_show', 'is_jupyter', 'is_reticulate']

def plot_args(kwargs=None, _debug=False, **defaults):
    """
    Process known plotting kwargs.

    This function handles arguments to `sim.plot()` and other plotting functions
    by splitting known kwargs among all the different aspects of the plot.

    Note: the kwargs supplied to the parent function should be supplied as the
    first argument of this function; keyword arguments to this function are treated
    as default values that will be overwritten by user-supplied values in `kwargs`.
    The argument "_debug" is used internally to print debugging output, but is
    not typically set by the user.

    Args:
        fig_kw (dict): passed to `sc.getrowscols()`, then `plt.subplots()` and `plt.figure()`
        plot_kw (dict): passed to `plt.plot()`
        data_kw (dict): passed to `plt.scatter()`, for plotting the data
        style_kw (dict): passed to `sc.options.with_style()`, for controlling the detailed plotting style
        **kwargs (dict): parsed among the above dictionaries

    Returns:
        A dict-of-dicts with plotting arguments, for use with subsequent plotting commands

    Valid kwarg arguments are:

        - fig: 'figsize', 'nrows', 'ncols', 'ratio', 'num', 'dpi', 'facecolor'
        - plot: 'alpha', 'c', 'lw', 'linewidth', 'marker', 'markersize', 'ms'
        - data: 'data_alpha', 'data_color', 'data_size'
        - style: 'font', 'fontsize', 'interactive'
        - return_fig: 'do_show', 'is_jupyter', 'is_reticulate'

    **Examples**:

        kw = ss.plot_args(kwargs, fig_kw=dict(figsize=(10,10)) # Explicit way to set figure size, passed to `plt.figure()` eventually
        kw = ss.plot_args(kwargs, figsize=(10,10)) # Shortcut since known keyword
    """
    suffix='_kw',
    _None = '<None>'
    kwargs = sc.mergedicts(defaults, kwargs) # Input arguments, e.g. ss.plot_args(kwargs, figsize=(8,6))
    kw = sc.objdict() # Output arguments
    for subtype,args in plotting_kw.items():
        if _debug: print('Subtype & args:', subtype, args)
        kw[subtype] = sc.objdict() # e.g. kw.fig

        # Handle kwargs, e.g. "figsize"
        if isinstance(args, list): # Ensure everything's a dict, although only kw.data is
            args = {k:k for k in args}
        for inkey,outkey in args.items():
            val = kwargs.pop(inkey, _None) # Handle None as a valid argument
            if _debug: # Just for debugging, since the logic is complex
                if inkey == outkey:
                    print(f'    {inkey} = {val}')
                else:
                    print(f'    {inkey} → {outkey} = {val}')
            if val is not _None:
                kw[subtype][outkey] = val

        # Handle dicts of kwargs, e.g. "fig_kw"
        subtype_dict = kwargs.pop(f'{subtype}{suffix}', None) # e.g. fig_kw
        if subtype_dict:
            if _debug: print('Subtype dict:', subtype_dict)
            kw[subtype].update(subtype_dict)

    # Expect everything has been converted
    if len(kwargs):
        valid = f'\n\nValid:\n{plotting_kw}\n'
        converted = f'\n\nConverted:\n{kw}\n'
        unconverted = f'\n\nUnconverted:\n{sc.newlinejoin(kwargs.keys())}'
        errormsg = f'Did not successfully convert all plotting keys:{valid}{converted}{unconverted}'
        raise sc.KeyNotFoundError(errormsg)

    if _debug: print('Final output:', kw)

    return kw


def match_result_keys(results, key, show_skipped=False, flattened=False):
    """ Ensure that the user-provided keys match available ones, and raise an exception if not """

    def normkey(key):
        """ Normalize the key: e.g. 'SIS.prevalence' becomes 'sis_prevalence' """
        return key.replace('.','_').lower()

    # Handle accessing subdicts
    if isinstance(key, str) and key in results and isinstance(results[key], dict): # Key matches a subdict, use that directly, e.g. sim.plot('sis')
        flat = results[key].flatten() # e.g. sim.results['sis']
        key = None # We've already used the key, so reset it
    else: # Main use case: flatten the dict, e.g. sim.plot()
        flat = results if flattened else results.flatten()

    # Configuration
    flat_orig = flat # Copy reference before we modify in place
    if not show_skipped: # Skip plots with auto_plot set to False
        for k in list(flat.keys()): # NB: can't call it "key", shadows argument
            res = flat[k]
            if isinstance(res, ss.Result) and not res.auto_plot:
                flat.pop(k)

    if key is not None:
        if isinstance(key, str):
            flat = {k:v for k,v in flat.items() if (normkey(key) in k)} # Will match e.g. 'SIS.prevalence' and 'sis_prevalence'
            if len(flat) != 1:
                errormsg = f'Key "{key}" not found; valid keys are:\n{sc.newlinejoin(flat_orig.keys())}'
                raise sc.KeyNotFoundError(errormsg)
        else:
            try:
                flat = {k.lower():flat[normkey(k)] for k in key}
            except sc.KeyNotFoundError as e:
                errormsg = f'Not all keys could be matched.\nAvailable keys:\n{sc.newlinejoin(flat_orig.keys())}\n\nYour keys:\n{sc.newlinejoin(key)}'
                raise sc.KeyNotFoundError(errormsg) from e

    return flat


def get_result_plot_label(res, show_module=None):
    """ Helper function for getting the label to plot for a result; not for the user """
    # Sanitize
    if show_module is None:
        show_module = 26 # Default maximum length
    elif isinstance(show_module, bool):
        if show_module is True:
            show_module = 999
        else:
            show_module = 0
    if not isinstance(show_module, int):
        errormsg = f'"show_module" must be a bool or int, not {show_module}'
        raise TypeError(errormsg)

    # Decide how/if to show the module
    if show_module == -1:
        label = res.full_label.replace(':', '\n')
    elif len(res.full_label) > show_module:
        label = sc.ifelse(res.label, res.name)
    else:
        label = res.full_label

    return label


def format_axes(ax, res, n_ticks=None, show_module=None):
    """ Standard formatting for axis results; not for the user """
    if n_ticks is None:
        n_ticks = (2,5)

    # Set y axis -- commas
    sc.commaticks(ax)

    # Set x axis -- date formatting
    if res.has_dates:
        locator = mpl.dates.AutoDateLocator(minticks=n_ticks[0], maxticks=n_ticks[1]) # Fewer ticks since lots of plots
        sc.dateformatter(ax, locator=locator)

    # Set the axes title
    label = ss.utils.get_result_plot_label(res, show_module)
    ax.set_title(label)
    return


def show(**kwargs):
    """ Shortcut for matplotlib.pyplot.show() """
    return plt.show(**kwargs)


def return_fig(fig, **kwargs):
    """ Do postprocessing on the figure: by default, don't return if in Jupyter, but show instead; not for the user """
    do_show = kwargs.pop('do_show', ss.options.show) # TODO: make this more consistent with other implementations, e.g. plot_args()
    is_jupyter = kwargs.pop('is_jupyter', ss.options.is_jupyter)
    is_reticulate = kwargs.pop('is_reticulate', ss.options.reticulate)
    if is_jupyter or is_reticulate:
        print(fig)
        if do_show != False:
            plt.show()
        return None
    else:
        if do_show:
            plt.show()
        return fig


"""
Networks based on DHS household data. One is static and the other evolves over
time, with females forming new households as they become pregnant.
"""
import numpy as np
import starsim as ss

# This has a significant impact on runtime, surprisingly
ss_float = ss.dtypes.float
_ = None

# Specify all externally visible functions this file defines; see also more definitions below
__all__ = ['HouseholdDHSNet', 'EvolvingHouseholdDHSNet']


class HouseholdDHSNet(ss.Network):
    """ Static DHS network

    A network class of static households derived from DHS data. There's no formation of new households. The network
    requires a DHS dataset. When this network is initialized it overrides the age and sex of all agents in the sim and
    assigns each agent a household ID. Use with caution if other modules depend upon or alter age and sex too. Because
    this network is static, it can have unexpected side effects such as growing or shrinking household sizes when used
    with aging enabled or with other demographics modules like Pregnancy.

    This network creates households by selecting a random household from the DHS data and setting the age and sex of
    agents to match, repeating until all agents have been assigned to a household. DHS data record ages in integer years
    but we typically don't want all agents to have exactly the same age, so we add a random fractional age to each
    agent.

    This network assumes only one mother per household. Births are automatically added to their mother's household network.

    Args:
        dhs_data (dataframe): A pandas dataframe containing DHS data with columns 'hh_id' and 'ages'

    Note: the dataframe should look something like this:

            hh_id                ages
       0        0          72, 17, 30
       1        1                  37
       2        2          13, 55, 36
       3        3  52, 13, 12, 64, 53
       4        4              30, 66

    **Example**:

        import numpy as np
        import sciris as sc
        import starsim as ss
        import starsim_examples as sse

        # Construct DHS data
        n = 1000
        hhid_id = np.arange(n)
        age_strings = []
        for i in range(n):
            household_size = np.random.randint(1,6)
            ages = np.random.randint(0, 80, household_size)
            age_strings.append(sc.strjoin(ages))
        dhs_data = sc.dataframe(hh_id=np.arange(n), ages=age_strings) # If real data are available: dhs_data = pd.read_csv(dhs_data_path)

        household_dhs = sse.HouseholdDHSNet(dhs_data=dhs_data)

        sim = ss.Sim(diseases='sis', networks=household_dhs)
        sim.run()
        sim.plot()
    """
    def __init__(self, pars=None, dhs_data=None, **kwargs):
        super().__init__(**kwargs)
        self.define_pars()
        self.update_pars(pars, **kwargs)
        self.dhs_data = dhs_data
        self.define_states(
            ss.FloatArr('household_ids'),
        )
        self.p_fractional_age = ss.uniform()
        return

    def init_pre(self, sim):
        super().init_pre(sim)
        if self.dhs_data is None:
            raise ValueError("Please provide DHS data to the HouseholdDHSNetwork.")
        return

    def init_post(self, add_pairs=True):
        super().init_post(add_pairs)

        # DHS age data is in integer years. This doesn't make sense for a simulation, so we add a random fractional age
        self.sim.people.age[:] = self.sim.people.age + self.p_fractional_age.rvs(self.sim.people.auids)
        return

    def add_pairs(self):
        """
        Generate contacts
        """
        ppl = self.sim.people
        pop_size = len(ppl)
        dhs = self.dhs_data

        # Initialize
        n_remaining = len(ppl)

        # Loop over the clusters
        cluster_id = -1
        p1 = []
        p2 = []
        while n_remaining > 0:
            cluster_id += 1 # Assign cluster id

            # Sample a household from the DHS data
            rand_row = np.random.choice(len(dhs))
            household_data = dhs.iloc[rand_row]
            age_data = household_data['ages']
            sex_data = None
            if 'sexes' in household_data.keys():
                sex_data = household_data['sexes']

            age_data = np.array([float(x) for x in age_data.split(', ')], dtype=float)  # Parse the ages
            cluster_size = len(age_data)  # Sample the cluster size

            if cluster_size > n_remaining:
                cluster_size = n_remaining

            # UIDS of people in this cluster
            cluster_uids = ss.uids((pop_size-n_remaining)+np.arange(cluster_size))

            # Set the ages and household_ids
            ppl.age[cluster_uids] = age_data[0:cluster_size]
            if sex_data is not None:
                ppl.sex[cluster_uids] = sex_data[0:cluster_size]
            self.household_ids[cluster_uids] = cluster_id

            # Add symmetric pairwise contacts in each cluster
            for i in cluster_uids:
                for j in cluster_uids:
                    if j > i:
                        p1.append(i)
                        p2.append(j)
            n_remaining -= cluster_size
        beta = np.ones(len(p1), dtype=ss_float)
        self.append(p1=p1, p2=p2, beta=beta)
        return

    # No updates on step, so pass
    def step(self):
        return


class EvolvingHouseholdDHSNet(HouseholdDHSNet):
    """
    Extends the HouseholdDHSNetwork by:

        1. Assigning one random female as head of each household
        2. Allowing females who are not current heads of household to move out and start their own household with a random male partner
        3. Adding new births into the household of the mother

    Args:
        dhs_data (dataframe): A pandas dataframe containing DHS data with columns 'hh_id' and 'ages'
        prob_move_out (float): The probability a female non-head of household moves out to start her own household with a randomly selected partner. Evaluated once at the beginning of each pregnancy.
        update_freq (int): The frequency of updates in timesteps, e.g. 7 for weekly if using daily timesteps, 1 if using weekly timesteps.
    """
    def __init__(self, pars=None, dhs_data=None, prob_move_out=_, update_freq=_, **kwargs):
        super().__init__(**kwargs)
        self.define_pars(
            prob_move_out = ss.bernoulli(p=0.7), # Probability that female moves out of her household when pregnant and not a head of household
            update_freq = 1,
        )
        self.update_pars(pars, **kwargs)
        self.dhs_data = dhs_data

        self.define_states(
            ss.BoolArr('fhoh', default='False'),  # An array for tracking the female head of household
            ss.FloatArr('ti_move_out_check', default='-inf'), # A time index for checking when to move out. Used to prevent multiple moveout checks per pregnancy.
        )
        return

    def init_pre(self, sim):
        super().init_pre(sim)
        ss.check_requires(self.sim, ['pregnancy'])
        return

    def add_pairs(self):
        # Generate the pairs, then select the fhoh for each household
        super().add_pairs()

        ppl = self.sim.people

        # Find a female head of household if one exists between the ages of 15 and 50
        # We just created new households, so the last cluster id is the total number of clusters
        for cluster_id in range(int(self.household_ids[-1])):
            cluster_uids = ss.uids(self.household_ids == cluster_id)

            female_uids = cluster_uids[
                ppl.female[cluster_uids] & (ppl.age[cluster_uids] >= 15) & (ppl.age[cluster_uids] <= 50)]
            if len(female_uids) > 0:
                fhoh = np.random.choice(a=female_uids)
                self.fhoh[ss.uids(fhoh)] = True


    def step(self):
        super().step()

        if np.mod(self.ti, self.pars.update_freq):
            return

        self.add_births()
        self.create_new_households()
        return

    def add_births(self):
        sim = self.sim
        birth_uids = ss.uids((sim.people.age >= 0) & (sim.people.age < self.pars.update_freq * self.sim.t.dt_year))
        if len(birth_uids) == 0:
            return 0

        mat_uids = sim.people.parent[birth_uids]
        # Could be very young people in the initial population or mother could have died
        keep = mat_uids != sim.people.parent.nan
        birth_uids = birth_uids[keep]
        mat_uids = mat_uids[keep]
        if len(birth_uids) == 0:
            return 0

        p1 = []
        p2 = []
        for new_uid, mat_uid in zip(birth_uids, mat_uids):
            hh_contacts = ss.uids(self.household_ids == self.household_ids[mat_uid])
            p1.append(hh_contacts)
            p2.append([new_uid] * len(hh_contacts))

        p1 = ss.uids.cat(p1)
        p2 = ss.uids.cat(p2)

        beta = np.ones(len(p1), dtype=ss.dtypes.float)
        self.append(p1=p1, p2=p2, beta=beta)

        self.household_ids[birth_uids] = self.household_ids[mat_uids]  # Assign HHID (after adding network edges)

        return len(birth_uids)

    def create_new_households(self):
        """
        Create new households. Find females that are pregnant and not a head of household.
        Move them and a randomly sampled male partner to a new household.
        Then fully connect those contacts.
        """
        ppl = self.sim.people
        potential_movers = ss.uids(~self.fhoh & ppl.pregnancy.pregnant & (self.ti_move_out_check <= self.sim.ti))
        moving_out = self.pars['prob_move_out'].filter(potential_movers)
        if len(moving_out) > 0:
            self.fhoh[moving_out] = True
            potential_partners = ss.uids(ppl.male & (ppl.age > 15) & (ppl.age < 50))
            partner_inds = np.random.permutation(len(potential_partners))[:len(moving_out)]
            partners = potential_partners[partner_inds]  # Faster than np.random.shuffle
            to_remove = ss.uids.cat([moving_out, partners])
            self.remove_uids(to_remove)  # Remove their old contacts from the HH network
            beta = np.ones(len(moving_out), dtype=ss.dtypes.float)
            self.append(p1=moving_out, p2=partners, beta=beta)  # Create a new household

            # Create new household IDs
            new_cids = np.nanmax(self.household_ids) + 1 + np.arange(len(moving_out))
            self.household_ids[moving_out] = new_cids
            self.household_ids[partners] = new_cids

        # Only consider move out once per pregnancy
        self.ti_move_out_check[potential_movers] = ppl.pregnancy.ti_postpartum[
            potential_movers]  # Update the time index for checking when to move out
        return


#!/usr/bin/env python3
import sys
from pathlib import Path

files = [Path(arg) for arg in sys.argv[1:]] if len(sys.argv) > 1 else Path('.').rglob('*.py')

for path in files:
    text = path.read_text()
    new_text = text.replace('ss.State', 'ss.BoolState')
    if new_text != text:
        path.write_text(new_text)
        print(f"Updated {path}") 


"""
Simulation and module timelines
"""
import sciris as sc
import numpy as np
import starsim as ss

__all__ = ['Timeline']


class Timeline:
    """
    Handle time vectors and sequencing ("timelines") for both simulations and modules.

    Each module can have its own time instance, in the case where the time vector
    is defined by absolute dates, these time vectors are by definition aligned. Otherwise
    they can be specified using dur objects which express relative times (they can be added
    to a date to get an absolute time)

    Args:
        start (str/int/float/`ss.date`/`ss.dur`): when the simulation/module starts, e.g. '2000', '2000-01-01', 2000, ss.date(2000), or ss.years(2000)
        stop (str/int/float/`ss.date`/`ss.dur`): when the simulation/module ends (note: if start is a date, stop must be too)
        dt (int/float/`ss.dur`): Simulation step size
        dur (int/float/`ss.dur`): If "stop" is not provided, run for this duration
        name (str): if provided, name the `Timeline` object
        init (bool): whether or not to immediately initialize the `Timeline` object (by default, yes if start and stop or start and dur are provided; otherwise no)
        sim (Sim): if provided, initialize the `Timeline` with this as the parent (i.e. populating missing values)

    The `Timeline` object, after initialization, has the following time vectors,
    each representing a different way of representing time:

    - `tvec`: ground truth simulation time, either as absolute `ss.date` instances, or relative `ss.dur` instances, e.g. `DateArray([<2021.01.01>, <2021.01.03>, <2021.01.05>, <2021.01.07>])`
    - `tivec`: the vector of time indices (`np.arange(len(tvec))`)
    - `timevec`: the "human-friendly" representation of `tvec`: same as `tvec` if using `ss.date`, else floats if using `ss.dur`
    - `yearvec`: time represented as floating-point years
    - `datevec`: time represented as `ss.date` instances
    - `relvec`: relative time, in the sim's time units

    The `Timeline` object also has the following attributes/methods:

    - `ti` (int): the current timestep
    - `npts` (int): the total number of timesteps
    - `now()` (`ss.date`/float/str): the current time, based on the timevec by default or a different vector if specified

    **Examples**:

        t1 = ss.Timeline(start=2000, stop=2020, dt=1.0)
        t2 = ss.Timeline(start='2021-01-01', stop='2021-04-04', dt=ss.days(2))
    """

    # Allowable time arguments
    time_args = ['start', 'stop', 'dt']
    _time_vecs = ['tvec', 'tivec', 'timevec', 'yearvec', 'datevec', 'relvec']

    def __init__(self, start=None, stop=None, dt=None, dur=None, name=None, init=None, sim=None):
        # Store inputs
        self.name = name
        self.start = start
        self.stop = stop
        self.dt = dt
        self.dur = dur

        # Set defaults
        self.default_type = ss.years # The default ss.dur type, e.g. ss.years; this is reset after all the parameters are reconciled
        self.default_year_start = 2000
        self.default_rel_start = 0
        self.default_start = None # This is set to year_start or rel_start depending on default_type (in reconcile_args())
        self.default_dur = 50
        self.default_dt = 1.0

        # Populated later
        self.ti = 0 # The time index, e.g. 0, 1, 2
        self.tvec    = None # The time vector for this instance in date or dur format
        self.tivec   = None # The time index vector
        self.timevec = None # The human-friendly time representation
        self.yearvec = None # Time vector as floating point years
        self.datevec = None # Time vector as date objects
        self.relvec  = None # Time vector in sim time units
        self.is_numeric = False # Whether all inputs provided are numeric (e.g. start=2000, stop=2010, dt=0.1)
        self.initialized = False # Call self.init(sim) to initialize the object

        # Decide whether to initialized: we're asked, a sim is provided, or arguments are supplied directly
        if init or sim or init is None and sum([x is not None for x in [start, stop, dur]]) >= 2:
            self.init(sim)
        return

    def __repr__(self):
        if self.initialized:
            return f'Timeline({self.start}-{self.stop}; dt={self.dt}; now={self.tvec[self.ti]}; ti={self.ti}/{len(self)-1})'
        else:
            return 'Timeline(uninitialized)>'

    def disp(self):
        return sc.pr(self)

    @property
    def npts(self):
        try:
            return self.tvec.shape[0]
        except:
            return 0

    def to_dict(self):
        """ Return a dictionary of all time vectors """
        out = sc.objdict()
        for key in self._time_vecs:
            out[key] = getattr(self, key)
        return out

    @property
    def dt_year(self):
        """ The timestep size in years """
        return self.dt.years

    @property
    def year(self):
        """ The current time in years """
        return self.now('year')

    def __bool__(self):
        """ Always truthy """
        return True

    def __len__(self):
        """ Length is the number of timepoints """
        return self.npts

    @property
    def ready(self):
        """ Check if all parameters are in place to be initialized """
        return not any([getattr(self, k) is None for k in self.time_args])

    @property
    def finished(self):
        """
        Check if the simulation is finished, i.e. we're at the last time point
        (note, this does not distinguish whether we are at the beginning or end
        of the last time point, so use with caution!)
        """
        return self.ti == self.npts-1

    @property
    def is_absolute(self):
        """
        Check whether the fundamental simulation unit is absolute

        A time vector is absolute if the start is a date rather than a dur
        A relative time vector can be made absolute by adding a date to it.
        """
        try:
            return isinstance(self.start, ss.date)
        except:
            return False

    def now(self, key=None):
        """
        Get the current simulation time

        Args:
            key (str): which type of time to get: "time" (default), "year", "date", "tvec", or "str"

        **Examples**:

            t = ss.Timeline(start='2021-01-01', stop='2022-02-02', dt='week')
            t.ti = 25
            t.now() # Returns <2021-06-25>
            t.now('date') # Returns <2021-06-25>
            t.now('year') # Returns 2021.479
            t.now('str') # Returns '2021-06-25'
        """
        # Preprocessing
        to_str = False
        if key in [None, 'none', 'str']: # All of these are the default
            if key == 'str':
                to_str = True
            key = 'time'
        if not isinstance(key, str):
            errormsg = f'Key must be a string, not {key}'
            raise TypeError(errormsg)
        key = key.removesuffix('vec') + 'vec' # Allow either e.g. 'yearvec' or 'year'

        # Get the right keyvec
        if key in self._time_vecs:
            vec = getattr(self, key)
        else:
            errormsg = f'Invalid key "{key}": must be one of {sc.strjoin(self._time_vecs)}'
            raise ValueError(errormsg)

        if 0 <= self.ti < len(vec): # Normal use case, we're in the middle of a sim
            now = vec[self.ti]
        else: # Special case, we are before or after the sim period
            now = self.tvec[0] + self.dt*self.ti
            if key == 'yearvec':
                now = float(now)

        if to_str:
            now = str(now)

        return now

    def update(self, pars=None, parent=None, reset=True, force=None, **kwargs):
        """ Reconcile different ways of supplying inputs """
        pars = sc.mergedicts(pars)
        stale = False

        for key in self.time_args:
            current_val = getattr(self, key, None)
            parent_val = getattr(parent, key, None)
            kw_val = kwargs.get(key)
            par_val = pars.get(key)

            # Special handling for dt: don't inherit dt if the units are different
            if key == 'dt':
                if isinstance(parent, Timeline):
                    if parent.unit != self.unit:
                        parent_val = 1.0

            if force is False: # Only update missing (None) values
                val = sc.ifelse(current_val, kw_val, par_val, parent_val)
            elif force is None: # Prioritize current value
                val = sc.ifelse(kw_val, par_val, current_val, parent_val)
            elif force is True: # Prioritize parent value
                val = sc.ifelse(kw_val, par_val, parent_val, current_val)
            else:
                errormsg = f'Invalid value {force} for force: must be False, None, or True'
                raise ValueError(errormsg)

            if val != current_val:
                setattr(self, key, val)
                stale = True

        if stale and reset and self.initialized:
            self.init()
        return

    def reconcile_args(self, sim=None):
        """ Reconcile the different options for the input parameters """

        if sim is not None:
            self.dt    = sc.ifelse(self.dt,    sim.t.dt,    sim.pars.dt)
            self.start = sc.ifelse(self.start, sim.t.start, sim.pars.start)
            self.stop  = sc.ifelse(self.stop,  sim.t.stop,  sim.pars.stop)
            if self.start is None or self.stop is None: # Only set dur if start or stop is not specified
                self.dur   = sc.ifelse(self.dur,   sim.t.dur,   sim.pars.dur)

        # Convert strings to other types, starting with dt
        if isinstance(self.dt, str): # e.g. dt='year'
            dur_class = ss.time.get_dur_class(self.dt)
            self.dt = dur_class(1)

        # Convert start and stop from strings to dates
        if isinstance(self.start, str):
            self.start = ss.date(self.start)
        if isinstance(self.stop, str):
            self.stop = ss.date(self.stop)

        # Check to see if any inputs were provided as durations: if so, reset the default type
        args = [self.dt, self.start, self.stop, self.dur]
        for arg in args:
            if isinstance(arg, ss.dur) and not isinstance(arg, ss.datedur):
                self.default_type = type(arg)
                break # Stop at the first one

        # Decide if the time provided is "datelike" (actual dates or a date-like starting year)
        args = [self.start, self.stop, self.dur, self.dt] # These are modified from the above
        cal_year_like = self.default_type.base == 'years' # The default default_type is ss.years(), so this is True
        use_dates = False
        for arg in args:
            if isinstance(arg, (ss.date, ss.datedur)):
                cal_year_like = True
                use_dates = True
                break # Dates take precedence, so stop the loop here
            elif ss.time.assume_cal_year(arg):
                cal_year_like = True # Use 2000 as the start, but do not convert to dates

        # Set the default start based on whether we have datelike inputs
        self.default_start = self.default_year_start if cal_year_like else self.default_rel_start # Sets to the start to 2000 or 0

        # Check to see if all inputs are numeric
        self.is_numeric = all(arg is None or sc.isnumber(arg) for arg in args) # All inputs are either None or a number # Note: not currently used, although could be for setting timevec defaults

        # Ensure dur is valid
        self.default_dur = self.default_type(self.default_dur) # TODO: should this be ss.datedur() if use_dates = True? If so can move to the if block below
        if sc.isnumber(self.dur):
            self.dur = self.default_type(self.dur)
        if not (self.dur is None or isinstance(self.dur, ss.dur)):
            errormsg = f'Timeline.dur must be a number, a dur object or None, not {self.dur}'
            raise TypeError(errormsg)

        # Ensure dt is valid
        if not isinstance(self.dt, ss.dur):
            if self.dt is None: # Very fancy code to set self.dt to 1
                self.dt = self.default_dt
            if sc.isnumber(self.dt): # We have the right default type by now, so use that to set the appropriate dt
                self.dt = self.default_type(self.dt) # Defaults to 1.0 if not

        # Convert start and stop from numbers to either durations or dates
        if use_dates:
            if self.start is not None: self.start = ss.date(self.start)  # Convert numbers, durations, etc. to dates
            if self.stop  is not None: self.stop  = ss.date(self.stop)
            self.default_start = ss.date(self.default_start) # e.g. ss.date('2000.01.01')
        else:
            if sc.isnumber(self.start): self.start = self.default_type(self.start) # Only convert numbers to durations
            if sc.isnumber(self.stop):  self.stop  = self.default_type(self.stop)
            self.default_start = self.default_type(self.default_start) # e.g. ss.years(2000)

        # Validate durations: dt and dur
        for attr in ['dt', 'dur']:
            val = getattr(self, attr)
            if not (val is None or isinstance(val, ss.dur)):
                errormsg = f'Failed to parse {attr} = {val}: expecting ss.dur or None, not {type(val)}'
                raise TypeError(errormsg)

        # Validate start and stop
        for attr in ['start', 'stop']:
            val = getattr(self, attr)
            if not (val is None or isinstance(val, (ss.date, ss.dur))):
                errormsg = f'Failed to parse {attr} = {val}: expecting ss.date, ss.dur, or None, not {type(val)}'
                raise TypeError(errormsg)

        # Now, figure out start, stop, and dur: by this point, any supplied values should be of the correct type (date or dur, not str)
        match (self.start, self.stop, self.dur):
            case (None, None, None): # e.g. ss.Sim()
                start = self.default_start # e.g. ss.years(2000)
                dur = self.default_dur # e.g. ss.years(50)
                stop = start + dur # e.g. ss.years(2050)

            case (start, None, None): # e.g. ss.Sim(start=2000) or ss.Sim(start=ss.years(2000) or ss.Sim(start='2000.1.1')
                dur = self.default_dur # e.g. ss.years(50)
                stop = start + dur # e.g. ss.years(2050) or ss.date(2050)

            case (None, stop, None): # e.g. ss.Sim(stop=20) or ss.Sim(stop=2020) or ss.Sim(stop=ss.date(2020))
                if isinstance(stop, ss.dur) and not ss.time.assume_cal_year(stop):  # e.g. stop of ss.years(20), start will be ss.years(0)
                    start = stop.__class__(value=0)
                    dur = stop - start
                else: # e.g. stop of ss.years(2040), start will be ss.years(1990)
                    dur = self.default_dur
                    start = stop - dur

            case (None, None, dur): # e.g. ss.Sim(dur=20)
                start = self.default_start # e.g. ss.years(2000)
                stop = start + dur # e.g. ss.years(2020)

            case (start, None, dur): # e.g. ss.Sim(start=0, dur=20)
                stop = start + dur # e.g. ss.years(20)

            case (None, stop, dur): # e.g. ss.Sim(stop=2040, dur=20) or ss.Sim(stop=ss.date(2040), dur=20)
                start = stop - dur # e.g. ss.years(2020) or ss.date(2020)

            case (start, stop, None): # e.g. ss.Sim(start=2010, stop=2030)
                dur = stop - start # e.g. ss.years(20) or ss.datedur(years=20) (actually usually days)

            case (start, stop, dur): # e.g. ss.Sim(start=2010, stop=2030, dur=50)
                if dur != stop - start: # This is fine unless they don't match
                    errormsg = f'You supplied {start = }, {stop = }, and {dur = }, but {dur = } ≠ stop - start = {stop - start}'
                    raise ValueError(errormsg)

            case _: # This should not occur since we matched all 8 cases above
                errormsg = f'Failed to match {self.start = }, {self.stop = }, and {self.dur = } to any known pattern. You can use numbers, strings, ss.date, or ss.dur objects.'
                raise ValueError(errormsg)

        # Additional validation
        start_type = type(start)
        stop_type = type(start)
        assert isinstance(start, (ss.date, ss.dur)), f'Start must be ss.date or ss.dur, not {start_type}'
        assert isinstance(stop, (ss.date, ss.dur)), f'Stop must be ss.date or ss.dur, not {stop_type}'
        assert isinstance(dur, ss.dur), f'Duration must be ss.dur, not {type(dur)}'
        assert start_type is stop_type, f'Start and stop must be the same type, not {start_type} and {stop_type}'
        assert start <= stop, f'Start must be before stop, not {start} and {stop}'
        if (stop - start) < self.dt:
            warnmsg = f'The difference between {start = } and {stop = } is less than dt = {self.dt}; no timesteps will be run.'
            ss.warn(warnmsg)

        # Store everything
        self.start = start
        self.stop = stop
        self.dur = dur

        return

    def init(self, sim=None, max_steps=20_000, force=False):
        """ Initialize all vectors """

        # Don't re-initialize if already initialized
        if self.initialized and not force:
            return self

        # Handle start, stop, dt, dur
        self.reconcile_args(sim)

        # If the sim is provided and matches the current object: copy from the sim
        tvkeys = ['start', 'stop', 'dt']
        if sim is not None: # Sim is provided
            if sim.t.initialized: # It's initialized
                if all([type(getattr(self, key)) == type(getattr(sim.t, key)) for key in tvkeys]): # Types match
                    if all([getattr(self, key) == getattr(sim.t, key) for key in tvkeys]): # Values match
                        for attr in self._time_vecs: # Copy the time vectors over
                            new = sc.dcp(getattr(sim.t, attr))
                            setattr(self, attr, new)
                        self.initialized = True
                        return self

        # We need to make the tvec (ss.dates/ss.durs) the yearvec (years), and the datevec (dates). However, we
        # need to decide which of these quantities to prioritise considering that the calendar dates
        # don't convert consistently into fractional years due to varying month/year lengths. We will
        # prioritise one or the other depending on what type of quantity the user has specified for start
        self.datevec = ss.date.arange(self.start, self.stop, self.dt, allow_zero=True)
        n_steps = len(self.datevec)
        if n_steps > max_steps and ss.options.warn_convert:
            warnmsg = f'You have specified start={self.start}, stop={self.stop}, and dt={self.dt}, which results in {n_steps:n} timesteps. '
            warnmsg += 'This is above the recommended maximum of {max_steps:n}, which is valid, but inadvisable. '
            warnmsg += 'Set ss.options.warn_convert = False to disable this warning.'
            ss.warn(warnmsg)

        if isinstance(self.dt, ss.datedur):
            if isinstance(self.start, ss.dur): # e.g. ss.Sim(start=ss.years(2000), dt=ss.datedur(months=1))
                self.tvec = ss.dur.arange(self.start, self.stop, self.dt)
            else: # e.g. ss.Sim(start=ss.date(2000), dt=ss.datedur(months=1))
                self.tvec = ss.date.arange(self.start, self.stop, self.dt)
            self.yearvec = np.array([x.years for x in self.tvec])

        else: # e.g. self.dt = ss.years, ss.days
            if isinstance(self.start, ss.dur): # Use durations
                start = self.start
                stop = self.stop
                dt = self.dt
                eps = 1e-6 # Avoid rounding errors
                if type(start) == type(stop) == type(dt) == self.default_type: # Everything matches: create the tvec, then convert to years
                    self.tvec = sc.inclusiverange(start.value, stop.value+eps, dt.value)
                    self.tvec = self.default_type(self.tvec)
                    self.yearvec = self.tvec.years
                else: # They don't match: convert to years, then create the tvec
                    decimals = 12 # Also for avoiding rounding errors
                    start = self.start.years
                    stop = self.stop.years
                    dt = self.dt.years
                    self.yearvec = np.round(start + sc.inclusiverange(0, stop-start+eps, dt), decimals=decimals)  # Subtracting off self.start.years in np.arange increases floating point precision for that part of the operation, reducing the impact of rounding
                    self.tvec = self.default_type(ss.years(self.yearvec))
            elif isinstance(self.start, ss.date): # e.g. ss.Sim(ss.date(2000))
                self.tvec = self.datevec
                self.yearvec = self.datevec.years
            else:
                errormsg = f'Unexpected start {self.start}: expecting ss.dur or ss.Date'
                raise TypeError(errormsg)


        # Finalize timevecs
        self.tvec = ss.DateArray(self.tvec) # Ensure tvec is a DateArray
        self.tivec = np.arange(self.npts) # Simple time indices
        self.timevec = self.tvec.to_human() # The most human-friendly version of the dates: dates if possible, else floats

        # Finally, create a vector of relative times in the sim's time unit (if available)
        try:
            date0 = sim.t.datevec[0]
            dt = sim.t.dt
        except:
            date0 = self.datevec[0]
            dt = self.dt

        # Get the class for dt, which we use as the unit for the durations
        if isinstance(dt, ss.dur):
            dur_class = type(dt)
            if dur_class == ss.datedur: # Don't use ss.datedur for dt, since we want something numeric
                dur_class = type(dt.to_dur())
        else:
            dur_class = self.default_type

        if isinstance(date0, ss.date): # This check is necessary for skipping this step for mock modules -- TODO, make tidier
            dur_vec = self.datevec - date0 # Convert this Timeline's datevec to dates relative to sim start date
        else: # If it's not, use years
            dur_vec = ss.years(self.yearvec - self.yearvec[0])
        dur_vec = dur_class(dur_vec) # Convert to the intended class
        self.relvec = dur_vec.to_array() # Only keep the numeric array

        # Check that everything is the same
        for k,v in self.to_dict().items():
            if len(v) != len(self):
                errormsg = f'Expected all vectors be the same length, but len({k})={len(v)} ≠ len(tvec)={len(self)}'
                raise ValueError(errormsg)

        # We're done, phew
        self.initialized = True
        return self



"""
Test Starsim features not covered by other test files
"""
import sciris as sc
import numpy as np
import starsim as ss
import starsim_examples as sse
import matplotlib.pyplot as plt
import pytest

sc.options.interactive = False # Assume not running interactively
# ss.options.warnings = 'error' # For additional debugging

small = 100
medium = 1000

# %% Define the tests

@sc.timer()
def test_people():
    sc.heading('Testing people object')

    # Base people contains only the states defined in base.base_states
    ppl = ss.People(small)
    del ppl

    # Possible to initialize people with extra states, e.g. a geolocation
    def geo_func(n):
        locs = [1,2,3]
        return np.random.choice(locs, n)
    extra_states = [
        ss.FloatArr('geolocation', default=geo_func),
    ]
    ppl = ss.People(small, extra_states=extra_states)

    # Possible to add a module to people outside a sim (not typical workflow)
    ppl.add_module(sse.HIV())

    return ppl


@sc.timer()
def test_microsim(do_plot=False):
    sc.heading('Test small HIV simulation')

    # Make HIV module
    hiv = sse.HIV()
    # Set beta. The first entry represents transmission risk from infected p1 -> susceptible p2
    # Need to be careful to get the ordering right. The set-up here assumes that in the simple
    # sexual  network, p1 is male and p2 is female. In the maternal network, p1=mothers, p2=babies.
    hiv.pars['beta'] = {'mf': [0.15, 0.10], 'maternal': [0.2, 0]}

    sim = ss.Sim(
        people=ss.People(small),
        networks=[ss.MFNet(), ss.MaternalNet()],
        demographics=ss.Pregnancy(),
        diseases=hiv,
        copy_inputs = False, # So we can reuse hiv
    )
    sim.init()
    sim.run()

    if do_plot:
        plt.figure()
        plt.plot(hiv.timevec, hiv.results.n_infected)
        plt.title('HIV number of infections')

    return sim


@sc.timer()
def test_ppl_construction(do_plot=False):
    sc.heading('Test making people and providing them to a sim')

    def init_debut(module, sim, uids):
        # Test setting the mean debut age by sex, 16 for men and 21 for women.
        loc = np.full(len(uids), 16)
        loc[sim.people.female[uids]] = 21
        return loc

    mf_pars = {
        'debut': ss.normal(loc=init_debut, scale=2),  # Age of debut can vary by using callable parameter values
    }
    sim_pars = {'networks': [ss.MFNet(**mf_pars)], 'n_agents': small}
    gon_pars = {'beta': {'mf': [0.08, 0.04]}}
    gon = sse.Gonorrhea(**gon_pars)

    sim = ss.Sim(pars=sim_pars, diseases=[gon])
    sim.init()
    sim.run()
    if do_plot:
        plt.figure()
        plt.plot(sim.timevec, sim.results.gonorrhea.n_infected)
        plt.title('Number of gonorrhea infections')

    return sim


@sc.timer()
def test_arrs():
    sc.heading('Testing Arr objects')
    o = sc.objdict()

    # Create a sim with only births
    pars = dict(n_agents=medium, diseases='sis', networks='random')
    p1 = sc.mergedicts(pars, birth_rate=ss.freqperyear(10))
    p2 = sc.mergedicts(pars, death_rate=ss.freqperyear(10))
    s1 = ss.Sim(pars=p1).run()
    s2 = ss.Sim(pars=p2).run()

    # Tests
    assert len(s1.people.auids) > len(s2.people.auids), 'Should have more people with births'
    assert np.array_equal(s1.people.age, s1.people.age.raw[s1.people.auids]), 'Different ways of indexing age should match'
    assert np.array_equal(s1.people.alive.uids, s1.people.auids), 'Active/alive agents should match'
    assert np.all(s2.people.alive), 'All agents should be alive when indexed like this'
    assert not np.all(s2.people.alive.raw), 'Some agents should not be alive when indexed like this'
    assert np.array_equal(~s1.people.female, s1.people.male), 'Definition of men does not match'
    assert isinstance(s1.people.age < 5, ss.BoolArr), 'Performing logical operations should return a BoolArr'
    assert np.array_equal(s1.people.age[s1.people.age < 5],s1.people.age[s1.people.age.values < 5])

    o.s1 = s1
    o.s2 = s2

    return o


@sc.timer()
def test_deepcopy():
    sc.heading('Testing deepcopy')
    s1 = ss.Sim(pars=dict(diseases='sir', networks=sse.EmbeddingNet()), n_agents=small)
    s1.init()

    s2 = sc.dcp(s1)

    s1.run()
    s2.run()
    s1.plot()
    s2.plot()

    ss.diff_sims(s1, s2, full=True)
    assert np.allclose(s1.summary[:], s2.summary[:], rtol=0, atol=0, equal_nan=True)

    return s1, s2


@sc.timer()
def test_deepcopy_until():
    sc.heading('Testing deepcopy with until')
    s1 = ss.Sim(diseases=ss.SIR(init_prev=0.1), networks='random', n_agents=small)
    s1.init()

    s1.run(until=ss.date(2005))

    s2 = sc.dcp(s1)

    s1.run()
    s2.run()
    s1.plot()
    s2.plot()

    assert np.allclose(s1.summary[:], s2.summary[:], rtol=0, atol=0, equal_nan=True)

    return s1, s2


@sc.timer()
def test_results():
    sc.heading('Testing results export and plotting')

    # Make a sim with 2 SIS models with varying units and dt
    d1 = ss.SIS(dt=ss.years(1/12), name='sis1')
    d2 = ss.SIS(dt=ss.years(0.5), name='sis2')
    sim = ss.Sim(diseases=[d1, d2], networks='random')

    # Run sim and pull out disease results
    sim.run()
    rs1 = sim.results.sis1
    rs2 = sim.results.sis2

    # Export a single result to a series or dataframe
    res = rs1.new_infections
    res_df = res.to_df()
    res_series = res.to_series()
    assert res[-1] == res_df.iloc[-1].value == res_series.iloc[-1]

    # Export results of a whole module to a dataframe
    dfs = rs1.to_df()
    assert res_df.value.sum() == dfs.cum_infections.values[-1] == sim.summary.sis1_cum_infections

    # Export resampled summary of results to dataframe
    dfy1 = rs1.to_df(resample='year')
    dfy2 = rs2.to_df(resample='5YE')
    assert dfs.new_infections.iloc[:12].sum() == dfy1.new_infections.iloc[0]
    assert rs2.n_susceptible[:2].mean() == dfy2.n_susceptible.iloc[0]  # Entries 0 and 1 represent 2000
    assert rs2.n_susceptible[2:12].mean() == dfy2.n_susceptible.iloc[1]  # Entries 2-12 correspond to 2001-2005

    # Export whole sim to unified annualized dataframe
    sim_df = sim.to_df(resample='year', use_years=True)
    assert sim_df.sis1_n_infected.values[0] == rs1.n_infected[:12].mean()
    assert sim_df.sis2_n_infected.values[0] == rs2.n_infected[:2].mean()

    # Plot
    res.plot()
    sim.results.sis1.plot()
    sim.results.sis2.plot()

    return sim


@sc.timer()
def test_check_requires():
    sc.heading('Testing check_requires')
    s1 = ss.Sim(diseases='sis', networks='random', n_agents=medium).init()
    ss.check_requires(s1, 'sis')
    ss.check_requires(s1, ss.SIS)
    with pytest.raises(AttributeError):
        ss.check_requires(s1, ss.SIR)
    return s1


@sc.timer()
def test_mock_objects():
    sc.heading('Testing mock objects')
    o = sc.objdict()

    μ = 5
    mock = 100
    atol = 0.1

    # Initializing Dists
    o.dist = ss.lognorm_ex(μ, 2, mock=mock)
    assert np.isclose(o.dist.rvs(1000).mean(), μ, atol=atol), f'Distribution did not have expected mean value of {μ}'

    # Initializing Arrs -- can use it, but cannot grow it since mock slots can't grow
    o.arr = ss.FloatArr('my_arr', default=o.dist, mock=mock)
    assert len(o.arr) == mock, 'Arr did not have expected length of {mock}'
    assert np.isclose(o.arr.mean(), μ, atol=atol), f'Arr did not have expected mean value of {μ}'

    # Initializing People
    o.ppl = ss.People(n_agents=mock, mock=True)
    assert o.ppl.age.mean() > 15, 'People did not have the expected age'

    # Initializing Modules
    o.mod = ss.SIS().init_mock()
    for i in range(5):
        o.mod.step()
        o.mod.set_outcomes(ss.uids([i])) # Manually "infect" people
        o.mod.update_results()

    return o


@sc.timer()
def test_custom_imports():
    sc.heading('Testing custom imports')
    o = sc.objdict()

    class MyDisease(ss.SIS):
        pass

    class MyNetwork(ss.RandomNet):
        pass

    my_modules = [MyDisease, MyNetwork]

    # Make both starsim_examples and custom modules searchable
    ss.register_modules(sse, my_modules)
    sim = ss.Sim(n_agents=1000, diseases=['hiv', 'mydisease'], networks='mynetwork')
    sim.run()

    return o


# %% Run as a script
if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)

    # Start timing
    T = sc.tic()

    # Run tests
    ppl   = test_people()
    sim1  = test_microsim(do_plot)
    sim2  = test_ppl_construction(do_plot)
    sims  = test_arrs()
    sims2 = test_deepcopy()
    sims3 = test_deepcopy_until()
    sim4  = test_results()
    sim5  = test_check_requires()
    objs  = test_mock_objects()
    mods  = test_custom_imports()

    sc.toc(T)
    plt.show()
    print('Done.')



#!/usr/bin/env python
"""
Manually add aliases to Starsim functions, so instead of starsim.analyzers.Analyzer, you can link via starsim.Analyzer (and therefore ss.Analyzer)
"""
print('Customizing aliases ...')
import sciris as sc

# Get everything in the Starsim namespace -- this is the only part that needs to be customized
import starsim
mod_name = 'starsim'
mod_items = dir(starsim)

# Load the current objects inventory
filename = 'objects.json'
json = sc.loadjson(filename)
items = json['items']
names = [item['name'] for item in items]
print(f'  Loaded {len(json["items"])} items')

# Collect duplicates
dups = []
for item in items:
    parts = item['name'].split('.')
    if len(parts) < 3 or parts[0] != mod_name:
        continue
    objname = parts[2] # e.g. 'Analyzer' from starsim.analyzers.Analyzer
    if objname in mod_items:
        remainder = '.'.join(parts[2:])
        alias = f'{mod_name}.{remainder}'
        if alias not in names:
            dup = sc.dcp(item)
            dup['name'] = alias
            dups.append(dup)

# Add them back into the JSON
items.extend(dups)

# Save the modified version
sc.savejson(filename, json)
print(f'  Saved {len(json["items"])} items')


"""
Define Ebola model.
Adapted from https://github.com/optimamodel/gavi-outbreaks/blob/main/stisim/gavi/ebola.py
Original version by@domdelport and @RomeshA
"""

import numpy as np
import starsim as ss

__all__ = ['Ebola']


class Ebola(ss.SIR):

    def __init__(self, pars=None, **kwargs):
        """ Initialize with parameters """
        super().__init__()
        self.define_pars(
            # Initial conditions and beta
            init_prev       = ss.bernoulli(p=0.005),
            beta            = ss.prob(1.0, ss.days(1)), # Placeholder value
            sev_factor      = 2.2,
            unburied_factor = 2.1,

            # Natural history parameters, all specified in days
            dur_exp2symp    = ss.lognorm_ex(mean=ss.days(12.7)), # Add source
            dur_symp2sev    = ss.lognorm_ex(mean=ss.days(6)), # Add source
            dur_sev2dead    = ss.lognorm_ex(mean=ss.days(1.5)), # Add source
            dur_dead2buried = ss.lognorm_ex(mean=ss.days(2)), # Add source
            dur_symp2rec    = ss.lognorm_ex(mean=ss.days(10)), # Add source
            dur_sev2rec     = ss.lognorm_ex(mean=ss.days(10.4)), # Add source
            p_sev           = ss.bernoulli(p=0.7), # Add source
            p_death         = ss.bernoulli(p=0.55), # Add source
            p_safe_bury     = ss.bernoulli(p=0.25), # Probability of a safe burial - should be linked to diagnoses
        )
        self.update_pars(pars, **kwargs)

        # Boolean states
        self.define_states(
            # SIR states are added automatically, here we add exposed, severe, and buried
            ss.BoolState('exposed', label='Exposed'),
            ss.BoolState('severe', label='Severe'),
            ss.BoolState('buried', label='Buried'),

            # Timepoint states
            ss.FloatArr('ti_exposed', label='Time of exposure'),
            ss.FloatArr('ti_severe', label='Time of severe symptoms'),
            ss.FloatArr('ti_buried', label='Time of burial'),
        )
        return

    def step_state(self):

        # Progress exposed -> infected
        ti = self.ti
        infected = (self.exposed & (self.ti_infected <= ti)).uids
        self.exposed[infected] = False
        self.infected[infected] = True

        # Progress infectious -> severe
        severe = (self.infected & (self.ti_severe <= ti)).uids
        self.severe[severe] = True

        # Progress infected -> recovered
        recovered = (self.infected & (self.ti_recovered <= ti)).uids
        self.infected[recovered] = False
        self.recovered[recovered] = True

        # Progress severe -> recovered
        recovered_sev = (self.severe & (self.ti_recovered <= ti)).uids
        self.severe[recovered_sev] = False
        self.recovered[recovered_sev] = True

        # Trigger deaths
        deaths = (self.ti_dead <= ti).uids
        if len(deaths):
            self.sim.people.request_death(deaths)

        # Progress dead -> buried
        buried = (self.ti_buried <= ti).uids
        self.buried[buried] = True

        return

    def set_prognoses(self, uids, sources=None):
        """ Set prognoses for those who get infected """
         # We don't want to call super().set_prognoses(), but we could also do ss.Disease.set_prognoses(self, uids, sources)
        if self.infection_log:
            self.infection_log.add_entries(uids, sources, self.now)

        ti = self.ti
        self.susceptible[uids] = False
        self.exposed[uids] = True
        self.ti_exposed[uids] = ti

        p = self.pars

        # Determine when exposed become infected
        self.ti_infected[uids] = ti + p.dur_exp2symp.rvs(uids)

        # Determine who progresses to sever and when
        sev_uids = p.p_sev.filter(uids)
        self.ti_severe[sev_uids] = self.ti_infected[sev_uids] + p.dur_symp2sev.rvs(sev_uids)

        # Determine who dies and who recovers and when
        dead_uids = p.p_death.filter(sev_uids)
        self.ti_dead[dead_uids] = self.ti_severe[dead_uids] + p.dur_sev2dead.rvs(dead_uids)
        rec_sev_uids = np.setdiff1d(sev_uids, dead_uids)
        self.ti_recovered[rec_sev_uids] = self.ti_severe[rec_sev_uids] + p.dur_sev2rec.rvs(rec_sev_uids)
        rec_symp_uids = np.setdiff1d(uids, sev_uids)
        self.ti_recovered[rec_symp_uids] = self.ti_infected[rec_symp_uids] + p.dur_symp2rec.rvs(rec_symp_uids)

        # Determine time of burial - either immediate (safe burials) or after a delay (unsafe)
        safe_buried = p.p_safe_bury.filter(dead_uids)
        unsafe_buried = np.setdiff1d(dead_uids, safe_buried)
        self.ti_buried[safe_buried] = self.ti_dead[safe_buried]
        self.ti_buried[unsafe_buried] = self.ti_dead[unsafe_buried] + p.dur_dead2buried.rvs(unsafe_buried)

        # Change rel_trans values
        self.rel_trans[self.infectious] = 1
        self.rel_trans[self.severe] = self.pars['sev_factor']  # Change for severe
        unburied_uids = ((self.ti_dead <= ti) & (self.ti_buried > ti)).uids
        self.rel_trans[unburied_uids] = self.pars['unburied_factor']  # Change for unburied
        return

    def step_die(self, uids):
        # Reset infected/recovered flags for dead agents
        for state in ['susceptible', 'exposed', 'infected', 'severe', 'recovered']:
            self.state_dict[state][uids] = False
        return



"""
Version and license information.
"""

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '3.0.3'
__versiondate__ = '2025-10-07'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2025 by IDM'



"""
Utilities for running in parallel
"""
import numpy as np
import sciris as sc
import starsim as ss

__all__ = ['MultiSim', 'single_run', 'multi_run', 'parallel']


class MultiSim:
    """
    Class for running multiple copies of a simulation.

    Args:
        sims (Sim/list): a single sim or a list of sims
        base_sim (Sim): the sim used for shared properties; if not supplied, the first of the sims provided
        label (str): the name of the multisim
        n_runs (int): if a single sim is provided, the number of replicates (default 4)
        initialize (bool): whether or not to initialize the sims (otherwise, initialize them during run)
        inplace (bool): whether to modify the sims in-place (default True); else return new sims
        debug (bool): if True, run in serial
        kwargs (dict): stored in run_args and passed to run()
    """
    def __init__(self, sims=None, base_sim=None, label=None, n_runs=4, initialize=False,
                 inplace=True, debug=False, **kwargs):
        # Handle inputs
        if base_sim is None:
            if isinstance(sims, ss.Sim):
                base_sim = sims
                sims = None
            elif isinstance(sims, list):
                if not sims:
                    errormsg = 'You must supply at least one sim to create a MultiSim'
                    raise ValueError(errormsg)
                base_sim = sims[0]
            else:
                errormsg = 'If base_sim is not supplied, sims must be either a single sim '
                errormsg += '(treated as base_sim) or a list of sims, not {type(sims)}'
                raise TypeError(errormsg)

        # Set properties
        self.sims = sims
        self.base_sim = base_sim
        self.label = base_sim.label if (label is None and base_sim is not None) else label
        self.run_args = sc.mergedicts(dict(n_runs=n_runs, inplace=inplace, debug=debug), kwargs)
        self.results = None
        self.summary = None
        self.which = None  # Whether the multisim is to be reduced, combined, etc.
        self.timer = sc.timer() # Create a timer

        # Optionally initialize
        if initialize:
            self.init_sims()

        return

    def __len__(self):
        """ The length of a MultiSim is how many sims it contains """
        try:
            return len(self.sims)
        except:
            return 0

    def __repr__(self):
        """ Return a brief description of a multisim; see multisim.disp() for the more detailed version. """
        try:
            labelstr = f'"{self.label}"; ' if self.label else ''
            string   = f'MultiSim({labelstr}n_sims: {len(self)}; base: {self.base_sim})'
        except Exception as E:
            string = sc.objectid(self)
            string += f'Warning, multisim appears to be malformed:\n{str(E)}'
        return string

    def brief(self):
        """ A single-line display of the MultiSim; same as print(multisim) """
        print(self)
        return

    def show(self, output=False):
        """
        Print a moderate length summary of the MultiSim. See also multisim.disp()
        (detailed output) and multisim.brief() (short output).

        Args:
            output (bool): if true, return a string instead of printing output

        **Example**:

            msim = ss.MultiSim(ss.demo(run=False), label='Example multisim')
            msim.run()
            msim.show() # Prints moderate length output
        """
        labelstr = f' "{self.label}"' if self.label else ''
        simlenstr = f'{len(self)}'
        string  = f'MultiSim{labelstr} summary:\n'
        string += f'  Number of sims: {simlenstr}\n'
        string += f'  Reduced/combined: {self.which}\n'
        string += f'  Base: {self.base_sim}\n'
        if self.sims:
            string += '  Sims:\n'
            for s,sim in enumerate(self.sims):
                string += f'    {s}: {sim}\n'
        if not output:
            print(string)
            return
        else:
            return string

    def disp(self):
        """ Display the full object """
        return sc.pr(self)

    def init_sims(self, **kwargs):
        """
        Initialize the sims
        """

        # Handle which sims to use
        if self.sims is None:
            sims = self.base_sim
        else:
            sims = self.sims

        # Initialize the sims but don't run them
        kwargs = sc.mergedicts(self.run_args, kwargs, {'do_run': False})  # Never run, that's the point!
        kwargs.pop('inplace', None)
        kwargs.pop('debug', None)
        self.sims = multi_run(sims, **kwargs)

        return

    def run(self, **kwargs):
        """
        Run the sims; see `ss.multi_run()` for additional arguments

        Args:
            n_runs (int): how many replicates of each sim to run (if a list of sims is not provided)
            inplace (bool): whether to modify the sims in place (otherwise return copies)
            kwargs (dict): passed to multi_run(); use run_args to pass arguments to sim.run()

        Returns:
            None (modifies MultiSim object in place)
        """

        # Handle which sims to use -- same as init_sims()
        if self.sims is None:
            sims = self.base_sim
        else:
            sims = self.sims

            # Handle missing labels
            for s, sim in enumerate(sims):
                if sim.label is None:
                    sim.label = f'Sim {s}'

        # Run
        self.timer.start()
        kwargs = sc.mergedicts(self.run_args, kwargs)
        inplace = kwargs.pop('inplace', True)
        debug = kwargs.pop('debug', False)
        if debug:
            kwargs.pop('n_runs', None)
            kwargs.pop('iterpars', None)
            kwargs.pop('parallel', None)
            run_sims = [single_run(sim, **kwargs) for sim in sims]
        else: # The next line does all the work!
            run_sims = multi_run(sims, **kwargs) # Output sims are copies due to the pickling during parallelization

        # Handle output
        if inplace and isinstance(self.sims, list) and len(run_sims) == len(self.sims): # Validation
            for old,new in zip(self.sims, run_sims):
                old.__dict__.update(new.__dict__) # Update the same object with the new results
        self.sims = run_sims # Just overwrite references
        self.timer.stop()

        return self

    def _has_orig_sim(self):
        """ Helper method for determining if an original base sim is present """
        return hasattr(self, 'orig_base_sim')

    def _rm_orig_sim(self, reset=False):
        """ Helper method for removing the original base sim, if present """
        if self._has_orig_sim():
            if reset:
                self.base_sim = self.orig_base_sim
            delattr(self, 'orig_base_sim')
        return

    def shrink(self, **kwargs):
        """
        Not to be confused with reduce(), this shrinks each sim in the msim;
        see sim.shrink() for more information.

        Args:
            kwargs (dict): passed to sim.shrink() for each sim
        """
        self.base_sim.shrink(**kwargs)
        self._rm_orig_sim()
        for sim in self.sims:
            sim.shrink(**kwargs)
        return

    def reset(self):
        """ Undo reduce() by resetting the base sim, which, and results """
        self._rm_orig_sim(reset=True)
        self.which = None
        self.results = None
        return

    def reduce(self, quantiles=None, use_mean=False, bounds=None, output=False):
        """
        Combine multiple sims into a single sim statistically: by default, use
        the median value and the 10th and 90th percentiles for the lower and upper
        bounds. If use_mean=True, then use the mean and ±2 standard deviations
        for lower and upper bounds.

        Args:
            quantiles (dict): the quantiles to use, e.g. [0.1, 0.9] or {'low : '0.1, 'high' : 0.9}
            use_mean (bool): whether to use the mean instead of the median
            bounds (float): if use_mean=True, the multiplier on the standard deviation for upper and lower bounds (default 2)
            output (bool): whether to return the "reduced" sim (in any case, modify the multisim in-place)

        **Example**:

            msim = ss.MultiSim(ss.Sim())
            msim.run()
            msim.reduce()
            msim.summarize()
        """
        if use_mean:
            if bounds is None:
                bounds = 2
        else:
            if quantiles is None:
                quantiles = {'low': 0.1, 'high': 0.9}
            if not isinstance(quantiles, dict):
                try:
                    quantiles = {'low': float(quantiles[0]), 'high': float(quantiles[1])}
                except Exception as E:
                    errormsg = (f'Could not figure out how to convert {quantiles} into a quantiles object:'
                                f' must be a dict with keys low, high or a 2-element array ({str(E)})')
                    raise ValueError(errormsg) from E

        # Store information on the sims
        n_runs = len(self)
        reduced_sim = sc.dcp(self.sims[0])
        reduced_sim.metadata = dict(parallelized=True, combined=False, n_runs=n_runs, quantiles=quantiles,
                                    use_mean=use_mean, bounds=bounds)  # Store how this was parallelized

        # Calculate the statistics
        raw = {}

        rflat = reduced_sim.results.flatten()
        rkeys = list(rflat.keys())
        length_mismatches = sc.ddict(int)
        for rkey in rkeys:
            raw[rkey] = np.full((len(rflat[rkey]), len(self.sims)), np.nan)
            for s, sim in enumerate(self.sims):
                flat = sim.results.flatten()
                this_raw = raw[rkey]
                this_flat = flat[rkey]
                l1 = this_raw.shape[0]
                l2 = this_flat.shape[0]
                if l1 == l2:
                    length = l1
                else:
                    length_mismatches[sim.label] += 1
                    length = min(l1, l2)
                this_raw[:length, s] = this_flat[:length]
        if length_mismatches:
            warnmsg = 'Sim results have mismatched lengths; results have been truncated but are not necessarily aligned. Mismatches:\n'
            for k,v in length_mismatches.items():
                warnmsg += f'{k}: {v} mismatched results\n'
            ss.warn(warnmsg)

        for rkey in rkeys:
            res = rflat[rkey]
            if use_mean:
                r_mean = np.mean(raw[rkey], axis=1)
                r_std = np.std(raw[rkey], axis=1)
                res[:] = r_mean
                res.low = r_mean - bounds * r_std
                res.high = r_mean + bounds * r_std
            else:
                res[:] = np.quantile(raw[rkey], q=0.5, axis=1)
                res.low = np.quantile(raw[rkey], q=quantiles['low'], axis=1)
                res.high = np.quantile(raw[rkey], q=quantiles['high'], axis=1)

        # Compute and store final results
        reduced_sim.summarize()
        self.orig_base_sim = self.base_sim
        self.base_sim = reduced_sim
        self.results = ss.Results('MultiSim').merge(rflat) # Create the dictionary and merge it
        self.summary = reduced_sim.summary
        self.which = 'reduced'

        if output:
            return self.base_sim
        else:
            return self

    def mean(self, bounds=None, **kwargs):
        """
        Alias for reduce(use_mean=True). See reduce() for full description.

        Args:
            bounds (float): multiplier on the standard deviation for the upper and lower bounds (default, 2)
            kwargs (dict): passed to reduce()
        """
        return self.reduce(use_mean=True, bounds=bounds, **kwargs)

    def median(self, quantiles=None, **kwargs):
        """
        Alias for reduce(use_mean=False). See reduce() for full description.

        Args:
            quantiles (list or dict): upper and lower quantiles (default, 0.1 and 0.9)
            kwargs (dict): passed to reduce()
        """
        return self.reduce(use_mean=False, quantiles=quantiles, **kwargs)

    def summarize(self, method='mean', quantiles=None, how='default'):
        """
        Summarize the simulations statistically.

        Args:
            method (str): one of 'mean' (default: [mean, 2*std]), 'median' ([median, min, max]), or 'all' (all results)
            quantiles (dict): if method='median', use these quantiles
            how (str): passed to sim.summarize()
        """

        # Compute the summaries
        summaries = []
        for sim in self.sims:
            summaries.append(sim.summarize(how=how))

        summary = sc.dcp(summaries[0]) # Use the first one as a template
        for k in summary.keys():
            arr = np.array([s[k] for s in summaries])
            if method == 'all':
                summary[k] = arr
            elif method == 'mean':
                summary[k] = sc.objdict({'mean':arr.mean(), 'std':arr.std(), 'sem':sc.sem(arr)})
            elif method == 'median':
                if quantiles is None:
                    quantiles = sc.objdict({'median':0.5, 'min':0, 'max':1, 'q25':0.25, 'q75':0.75})
                elif isinstance(quantiles, list):
                    quantiles = {q:q for q in quantiles}
                summary[k] = {q:v for q,v in zip(quantiles, np.quantile(arr, quantiles))}

        self.summary = summary # Could reconcile with reduce()'s summary

        return summary

    def plot(self, key=None, fig=None, legend=True, **kwargs):
        """
        Plot all results in the MultiSim object.

        If the MultiSim object has been reduced (i.e. mean or median), then plot
        the best value and uncertainty bound. Otherwise, plot individual sims.

        Args:
            key (str): the results key to plot (by default, all)
            fig (Figure): if provided, plot results into an existing figure
            fig_kw (dict): passed to `sc.getrowscols()`, then `plt.subplots()` and `plt.figure()`
            plot_kw (dict): passed to `plt.plot()`
            data_kw (dict): passed to `plt.scatter()`, for plotting the data
            style_kw (dict): passed to `sc.options.with_style()`, for controlling the detailed plotting style
            fill_kw (dict): passed to `plt.fill_between()`
            legend_kw (dict): passed to `plt.legend()`
            legend (bool): whether to show the legend
            **kwargs (dict): known arguments (e.g. figsize, font) split between the above dicts; see `ss.plot_args()` for all valid options
        """
        # Has not been reduced yet, plot individual sim
        if self.which is None:
            fig = None
            res_keys = None
            kw = ss.plot_args(kwargs)
            alpha = kw.plot.get('alpha', 0.7 if len(self) < 5 else 0.5) # Set default alpha
            if key is None: # Set keys
                for sim in self.sims:
                    sim_keys = set(sim.results.flatten(only_auto=True).keys()) # Check if keys match for auto-plotting results
                    if res_keys is None:
                        res_keys = sim_keys
                    else:
                        if res_keys != sim_keys: # TODO: would be good to fix and plot all, but hard with sim.plot()
                            missing = res_keys - sim_keys
                            extra = sim_keys - res_keys
                            extratxt = f'\nExtra: {sc.strjoin(extra)}'
                            missingtxt = f'\nMissing: {sc.strjoin(missing)}'
                            warnmsg = f'Sim "{sim.label}" has different results keys:{extratxt}{missingtxt}\nResults may not plot correctly (i.e. axes titles may not be correct for all sims)'
                            ss.warn(warnmsg)
            for sim in self.sims: # Actually plot
                fig = sim.plot(key=key, fig=fig, alpha=alpha, is_jupyter=False, do_show=False, **kwargs)
            if legend:
                leg = None
                shape = getattr(fig, '_subplots_shape', None) # Sciris-generated figure
                if shape: # If we have empty space on the bottom right, put the legend there
                    n = np.prod(shape)
                    if len(fig.axes) != n: # Bottom-right axes is empty
                        ax = fig.add_subplot(shape[0], shape[1], n)
                        leg = sc.movelegend(fig.axes[0], ax, **kw.legend)
                if leg is None: # Otherwise, just put it in the last axes anyway
                    fig.axes[-1].legend(**kw.legend)

        # Has been reduced, plot with uncertainty bounds
        else:
            # Get arguments
            n_ticks = kwargs.pop('n_ticks', None)
            show_module = kwargs.pop('show_module', None)
            show_skipped = kwargs.pop('show_skipped', None)

            # Figure out the flat structure of results to plot
            flat = ss.utils.match_result_keys(self.results, key, show_skipped=show_skipped, flattened=True)

            # Set figure size
            n_cols,_ = sc.getrowscols(len(flat))
            default_figsize = np.array([8, 6])
            figsize_factor = np.clip((n_cols-3)/6+1, 1, 1.5) # Scale the default figure size based on the number of rows and columns
            figsize = default_figsize*figsize_factor
            kw = ss.plot_args(kwargs, figsize=figsize, alpha=0.8, fill_alpha=0.2, lw=2)

            # Get ready to plot
            with ss.style(**kw.style):
                if fig is None:
                    fig, axs = sc.getrowscols(len(flat), make=True, **kw.fig)
                else:
                    axs = sc.toarray(fig.axes)

                # Do the plotting
                for ax, (key, res) in zip(axs.flatten(), flat.items()):
                    ax.fill_between(res.timevec, res.low, res.high, **kw.fill)
                    ax.plot(res.timevec, res, **kw.plot)
                    ss.utils.format_axes(ax, res, n_ticks, show_module)

        return ss.return_fig(fig, **kw.return_fig)


def single_run(sim, ind=0, reseed=True, shrink=True, run_args=None, sim_args=None,
               verbose=None, do_run=True, **kwargs):
    """
    Convenience function to perform a single simulation run. Mostly used for
    parallelization, but can also be used directly.

    Args:
        sim         (Sim)   : the sim instance to be run
        ind         (int)   : the index of this sim
        reseed      (bool)  : whether to generate a fresh seed for each run
        noise       (float) : the amount of noise to add to each run
        noisepar    (str)   : the name of the parameter to add noise to
        shrink      (bool)  : whether to shrink the sim after the sim run
        run_args    (dict)  : arguments passed to sim.run()
        sim_args    (dict)  : extra parameters to pass to the sim, e.g. 'n_infected'
        verbose     (int)   : detail to print
        do_run      (bool)  : whether to actually run the sim (if not, just initialize it)
        kwargs      (dict)  : also passed to the sim

    Returns:
        sim (Sim): a single sim object with results

    **Example**:

        import starsim as ss
        sim = ss.Sim() # Create a default simulation
        sim = ss.single_run(sim) # Run it, equivalent(ish) to sim.run()
    """

    # Set sim and run arguments
    sim_args = sc.mergedicts(sim_args, kwargs)
    run_args = sc.mergedicts({'verbose': verbose}, run_args)
    if verbose is None:
        verbose = sim.pars['verbose']

    if not sim.label:
        sim.label = f'Sim {ind}'

    if reseed:
        sim.pars['rand_seed'] += ind  # Reset the seed, otherwise no point of parallel runs

    # Handle additional arguments
    for key, val in sim_args.items():
        if key in sim.pars.keys():
            if verbose >= 1:
                print(f'Setting key {key} from {sim[key]} to {val}')
            sim.pars[key] = val
        else:
            raise sc.KeyNotFoundError(f'Could not set key {key}: not a valid parameter name')

    # Run
    if do_run:
        sim.run(**run_args)

    # Shrink the sim to save memory
    if shrink:
        sim.shrink(die=False)

    return sim


def multi_run(sim, n_runs=4, reseed=None, iterpars=None, shrink=None, run_args=None, sim_args=None,
              par_args=None, do_run=True, parallel=True, n_cpus=None, verbose=None, **kwargs):
    """
    For running multiple sims in parallel. If the first argument is a list of sims
    rather than a single sim, exactly these will be run and most other arguments
    will be ignored.

    Note: `n_cpus=1` is *not* the same thing as setting `parallel=False`. The former
    still uses the parallelization methods (just on a single core), while the latter
    simply runs in a loop.

    Args:
        sim         (Sim/list): the sim instance to be run, or a list of sims.
        n_runs      (int)   : the number of parallel runs
        reseed      (bool)  : whether or not to generate a fresh seed for each run (default: true for single, false for list of sims)
        iterpars    (dict)  : any other parameters to iterate over the runs; see sc.parallelize() for syntax
        shrink      (bool)  : whether to shrink the sim after the sim run
        run_args    (dict)  : arguments passed to sim.run()
        sim_args    (dict)  : extra parameters to pass to the sim
        par_args    (dict)  : arguments passed to sc.parallelize()
        do_run      (bool)  : whether to actually run the sim (if not, just initialize it)
        parallel    (bool)  : whether to run in parallel using multiprocessing (else, just run in a loop)
        n_cpus      (int)   : the number of CPUs to run on (if blank, set automatically; otherwise, passed to par_args, and use all cores)
        verbose     (int)   : detail to print
        kwargs      (dict)  : also passed to the sim

    Returns:
        If combine is True, a single sim object with the combined results from each sim.
        Otherwise, a list of sim objects (default).

    **Example**:

        import starsim as ss
        sim = ss.Sim()
        sims = ss.multi_run(sim, n_runs=6)
    """

    # Handle inputs
    sim_args = sc.mergedicts(sim_args, kwargs)  # Handle blank
    par_args = sc.mergedicts({'ncpus': n_cpus}, par_args)  # Handle blank

    # Handle iterpars
    if iterpars is None:
        iterpars = {}
    else:
        n_runs = None  # Reset and get from length of dict instead
        for key, val in iterpars.items():
            new_n = len(val)
            if n_runs is not None and new_n != n_runs:
                raise ValueError(f'Each entry in iterpars must have the same length, not {n_runs} and {len(val)}')
            else:
                n_runs = new_n

    # Run the sims
    if isinstance(sim, ss.Sim):  # One sim
        if reseed is None: reseed = True
        iterkwargs = dict(ind=np.arange(n_runs))
        iterkwargs.update(iterpars)
        kwargs = dict(sim=sim, reseed=reseed, verbose=verbose, shrink=shrink,
                      sim_args=sim_args, run_args=run_args, do_run=do_run)
    elif isinstance(sim, list):  # List of sims
        if reseed is None: reseed = False
        iterkwargs = dict(sim=sim, ind=np.arange(len(sim)))
        kwargs = dict(reseed=reseed, verbose=verbose, shrink=shrink, sim_args=sim_args, run_args=run_args,
                      do_run=do_run)
    else:
        errormsg = f'Must be Sim object or list, not {type(sim)}'
        raise TypeError(errormsg)

    # Actually run
    if parallel:
        sims = sc.parallelize(single_run, iterkwargs=iterkwargs, kwargs=kwargs, **par_args)  # Run in parallel
    else:  # Run in serial, not in parallel
        sims = []
        n_sims = len(list(iterkwargs.values())[0])  # Must have length >=1 and all entries must be the same length
        for s in range(n_sims):
            this_iter = {k: v[s] for k, v in iterkwargs.items()}  # Pull out items specific to this iteration
            this_iter.update(kwargs)  # Merge with the kwargs
            this_iter['sim'] = this_iter['sim'].copy()  # Ensure we have a fresh sim; this happens implicitly on pickling with multiprocessing
            sim = single_run(**this_iter)  # Run in series
            sims.append(sim)

    return sims


def parallel(*args, **kwargs):
    """
    A shortcut to `ss.MultiSim()`, allowing the quick running of multiple simulations
    at once.

    Args:
        args (list): The simulations to run
        kwargs (dict): passed to multi_run()

    Returns:
        A run MultiSim object.

    **Examples**:

        s1 = ss.Sim(n_agents=1000, label='Small', diseases='sis', networks='random')
        s2 = ss.Sim(n_agents=2000, label='Large', diseases='sis', networks='random')
        ss.parallel(s1, s2).plot()
        msim = ss.parallel([s1, s2], shrink=False)
    """
    sims = sc.mergelists(*args)
    msim = MultiSim(sims=sims, **kwargs)
    msim.run()
    return msim



import starsim as ss

class MySIR:
    def __init__(self):
        super().__init__()
        self.define_states(
            ss.FloatArr('my_attr'),
            ss.BoolState('mystate', label='My new state'),
        )
        return 


#!/usr/bin/env python3
"""
Migrate v2 unit arguments to v3 dt arguments.

Rules:
- If dt is not defined or dt=1: change unit=<x> to dt=<x>
- If dt=<y>, change unit=<x>, dt=<y> to dt=ss.<x>(<y>)

Examples:
- ss.SIS(unit='day', dt=1.0) -> ss.SIS(dt=ss.days(1))
- ss.Births(unit='year', dt=0.25) -> ss.Births(dt=ss.years(0.25))
- ss.RandomNet(unit='week') -> ss.RandomNet(dt='week')
- ss.Sim(unit='day', dt=2) -> ss.Sim(dt=ss.days(2))
"""
import re
import sys
from pathlib import Path

unit_map = {
    'day': 'days',
    'week': 'weeks',
    'month': 'months',
    'year': 'years',
}

# Choose input files
files = [Path(arg) for arg in sys.argv[1:]] if len(sys.argv) > 1 else Path('.').rglob('*.py')

# Handles both orderings: unit then dt, or dt then unit
pattern_unit_dt = re.compile(r'unit\s*=\s*[\'"](\w+)[\'"]\s*,\s*dt\s*=\s*([\d\.]+)')
pattern_dt_unit = re.compile(r'dt\s*=\s*([\d\.]+)\s*,\s*unit\s*=\s*[\'"](\w+)[\'"]')
pattern_unit_only = re.compile(r'\bunit\s*=\s*[\'"](\w+)[\'"]') # Standalone unit

for path in files:
    text = path.read_text()
    original = text

    # Replace unit='x', dt=y → dt=ss.xs(y)
    def repl_unit_dt(m):
        unit, dt = m.group(1), m.group(2)
        ss_unit = unit_map.get(unit, unit + 's')
        return f'dt=ss.{ss_unit}({dt})'

    # Replace dt=y, unit='x' → dt=ss.xs(y)
    def repl_dt_unit(m):
        dt, unit = m.group(1), m.group(2)
        ss_unit = unit_map.get(unit, unit + 's')
        return f'dt=ss.{ss_unit}({dt})'

    # Replace unit='x' → dt='x' (assumes dt=1)
    def repl_unit_only(m):
        unit = m.group(1)
        return f"dt='{unit}'"

    text = pattern_unit_dt.sub(repl_unit_dt, text)
    text = pattern_dt_unit.sub(repl_dt_unit, text)
    text = pattern_unit_only.sub(repl_unit_only, text)

    if text != original:
        path.write_text(text)
        print(f'Updated {path}')


#!/usr/bin/env python3
"""
Run this script to save or check the multisim baseline.

You can supply command line arguments for "save" and "n_runs" as integers, e.g.
    ./multisim_baseline.py 0 20
"""

import sys
import test_baselines

save = False
n_runs = 10

if len(sys.argv) > 1:
    save = int(sys.argv[1])
    assert save in [0,1] # Avoid confusing with n_runs
if len(sys.argv) > 2:
    n_runs = int(sys.argv[2])

test_baselines.multisim_baseline(save, n_runs)
